import html
import sys
import traceback
from collections import defaultdict
from datetime import datetime
from io import StringIO

from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtCore import Qt
from PyQt5.QtGui import QCloseEvent, QMouseEvent
from PyQt5.QtWidgets import QMainWindow, QWidget

from dac.core.thread import ThreadWorker


class MainWindowBase(QMainWindow):
    def __init__(self) -> None:
        super().__init__()

        self.setWindowTitle("DAC Base Window")
        self.resize(1024, 768)

        self._thread_pool = QtCore.QThreadPool.globalInstance()
        self._progress_widget = ProgressWidget4Threads(self)

        self._settings = defaultdict(bool)

        self._log_widget: QtWidgets.QPlainTextEdit = None
        self._ipy_widget: QtWidgets.QWidget = None

    def _create_ui(self):
        self._log_widget = QtWidgets.QPlainTextEdit(parent=self) # the log Level selection?
        self._log_widget.appendHtml(f"<b>The log output:</b> @ {datetime.now():%Y-%m-%d} <br/>")
        self._log_widget.setReadOnly(True)
        self._log_widget.setLineWrapMode(QtWidgets.QPlainTextEdit.NoWrap)
        self._log_widget.hide()

    def _create_menu(self):
        menubar = self.menuBar()
        self._dac_menu = tool_menu = menubar.addMenu("&Tool")

        tool_menu.addSeparator()
        tool_menu.addAction("Toggle log output", self.action_toggle_log_widget, shortcut=Qt.CTRL+Qt.Key_L)
        tool_menu.addAction("Toggle IPyConsole", self.action_toggle_ipy_widget, shortcut=Qt.CTRL+Qt.Key_I)
        tool_menu.addSeparator()
        no_thread_action = QtWidgets.QAction("No threading", tool_menu)
        no_thread_action.setCheckable(True)
        no_thread_action.triggered.connect(lambda: self.action_toggle_setting("no_thread"))
        tool_menu.addAction(no_thread_action)

    def _create_status(self):
        status = DacStatusBar(self)
        self.setStatusBar(status)
        status.addPermanentWidget(self._progress_widget)

    def start_thread_worker(self, worker: ThreadWorker):
        worker.signals.message.connect(self.message)
        worker.signals.error.connect(self.excepthook)
        self._progress_widget.add_worker(worker)
        if self._settings["no_thread"]:
            worker.run()
        else:
            self._thread_pool.start(worker)

    def message(self, msg, log=True):
        self.statusBar().showMessage(msg, 3000)
        if log:
            self._log_widget.appendPlainText(f"{datetime.now():%H:%M:%S} - {msg}")

    def _action_resize_log_widget(self):
        h = self.height() - 60
        w = int(self.width() // 2.5)
        self._log_widget.setGeometry(self.width()-20-w, 30, w, h)

    def action_toggle_log_widget(self):
        if self._log_widget.isVisible():
            self._log_widget.hide()
        else:
            self._action_resize_log_widget()
            self._log_widget.show()
            self._log_widget.horizontalScrollBar().setValue(0)
            self._log_widget.raise_()

    def _action_resize_ipy_widget(self):
        if self._ipy_widget is None:
            return
        h = self.height() - 60
        w = int(self.width() // 2.5)
        self._ipy_widget.setGeometry(20, 30, w, h)

    def action_toggle_ipy_widget(self, **kwargs):
        if self._ipy_widget is None:
            try:
                from qtconsole.inprocess import \
                    QtInProcessKernelManager  # inprocess kernel can push variable
                from qtconsole.rich_jupyter_widget import RichJupyterWidget
            except ImportError:
                self.message("IPython console not available", log=False)
                return

            kernel_manager = QtInProcessKernelManager()
            kernel_manager.start_kernel()
            kernel = kernel_manager.kernel

            # if not hasattr(kernel, "io_loop"): # https://github.com/ipython/ipykernel/issues/319
            #     import ipykernel.kernelbase
            #     ipykernel.kernelbase.Kernel.start(kernel)
            # # "io_loop" issue got resolved in new version?

            kernel_client = kernel_manager.client()
            kernel_client.start_channels()

            ipython_widget = RichJupyterWidget(parent=self)
            ipython_widget.kernel_manager = kernel_manager
            ipython_widget.kernel_client = kernel_client

            ipython_widget.exit_requested.connect(lambda: ipython_widget.hide())

            self._ipy_widget = ipython_widget
        else:
            ipython_widget: RichJupyterWidget = self._ipy_widget

        if ipython_widget.isVisible() and not kwargs:
            ipython_widget.hide()
        else:
            kernel = ipython_widget.kernel_manager.kernel
            kernel.shell.push(kwargs)
            self._action_resize_ipy_widget()
            ipython_widget.show()
            ipython_widget.raise_()

    def action_toggle_setting(self, key):
        self._settings[key] = not self._settings[key]

    def resizeEvent(self, a0: QtGui.QResizeEvent) -> None:
        if self._log_widget.isVisible():
            self._action_resize_log_widget()
        if self._ipy_widget is not None and self._ipy_widget.isVisible():
            self._action_resize_ipy_widget()
        return super().resizeEvent(a0)

    def excepthook(self, etype, evalue, tracebackobj):
        self._log_widget.appendHtml(f"<br/><b><font color='red'>{etype.__name__}:</font></b> {evalue}")
        info_stream = StringIO()
        traceback.print_tb(tracebackobj, file=info_stream)
        info_stream.seek(0)
        info_str = info_stream.read()
        escaped_str = html.escape(info_str).replace('\n', '<br/>').replace(' ', '&nbsp;')
        self._log_widget.appendHtml(f"<div style='font-family:Consolas'>{escaped_str}</div>")

        self.message("Error occurred, check in log output <Ctrl-L>", log=False)

    def show(self) -> None:
        sys.excepthook = self.excepthook
        return super().show()
    
    def closeEvent(self, a0: QCloseEvent) -> None:
        # TODO: kill the threads in threadpool
        if self._ipy_widget is not None:
            self._ipy_widget.kernel_client.stop_channels()
            self._ipy_widget.kernel_manager.shutdown_kernel()
        return super().closeEvent(a0)

class ProgressBundle(QtWidgets.QWidget):
    def __init__(self, caption):
        super().__init__()
        layout = QtWidgets.QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        
        self._progressbar = progress_bar = QtWidgets.QProgressBar()
        progress_bar.setTextVisible(False)
        progress_bar.setMaximum(0)
        progress_bar.setFixedHeight(6)
        self._caption = caption
        self._label = label = QtWidgets.QLabel("<b style='color:orange;'>(Hold)</b> " + caption)
        layout.addWidget(label)
        layout.addWidget(progress_bar)

    def progress(self, i, n):
        self._progressbar.setMaximum(n)
        self._progressbar.setValue(i)

    def started(self):
        self._label.setText(self._caption)

    # TODO: dbl-click to cancel thread

class ProgressWidget4Threads(QtWidgets.QWidget):
    def __init__(self, parent) -> None:
        super().__init__(parent=parent)
        self._layout = QtWidgets.QHBoxLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self.setMinimumHeight(28)

    def add_worker(self, worker: ThreadWorker):
        progress_widget = ProgressBundle(worker.caption)
        worker.signals.progress.connect(progress_widget.progress)
        worker.signals.started.connect(progress_widget.started)
        def finished():
            self._layout.removeWidget(progress_widget)
        worker.signals.finished.connect(finished)
        self._layout.addWidget(progress_widget)
        # the original idea was to automatically switch among progress with one progressbar

class DacStatusBar(QtWidgets.QStatusBar):
    def mouseDoubleClickEvent(self, a0: QMouseEvent) -> None:
        self.parentWidget().action_toggle_log_widget()
        return super().mouseDoubleClickEvent(a0)
    

    
if __name__=="__main__":
    app = QtWidgets.QApplication(sys.argv)
    win = MainWindowBase()

    win._create_ui()
    win._create_menu()
    win._create_status()

    win.show()
    app.exit(app.exec())


import numpy as np
from dac.core.data import DataBase
from . import BinMethod, AverageType
from ..timedata import TimeData
from collections import namedtuple

class ProcessPackage: # bundle channels and ref_channel
    ...

class DataBins(DataBase):
    def __init__(self, name: str = None, uuid: str = None, y: np.ndarray=None, y_unit: str = "-") -> None:
        super().__init__(name, uuid)

        self.y = y if y is not None else np.array([])
        self.y_unit = y_unit
        self._method = BinMethod.Mean

class FreqDomainData(DataBase):
    def __init__(self, name: str = None, uuid: str = None, y: np.ndarray=None, df: float=1, y_unit: str="-") -> None:
        super().__init__(name, uuid)
    
        self.y = y if y is not None else np.array([]) # complex number
        self.y_unit = y_unit
        self.df = df

    @property
    def x(self):
        return np.arange(self.lines) * self.df
    
    @property
    def f(self):
        return self.x

    @property
    def lines(self):
        return len(self.y)

    @property
    def phase(self):
        return np.angle(self.y, deg=True)

    @property
    def amplitude(self):
        return np.abs(self.y)
    
    def remove_spec(self, bands: list[tuple[float, float]]):
        y = self.y.copy()
        x = self.x

        for ffrom, fto in bands:
            b = np.all([ffrom<=x, x<=fto], axis=0)
            y[b] = 0

        return FreqDomainData(
            name=f"{self.name}-FiltF",
            y=y,
            df=self.df,
            y_unit=self.y_unit
        )
    
    def keep_spec(self, bands: list[tuple[float, float]]):
        y = np.zeros_like(self.y)
        x = self.x

        for ffrom, fto in bands:
            b = np.all([ffrom<=x, x<=fto], axis=0)
            y[b] = self.y[b]

        return FreqDomainData(
            name=f"{self.name}-ExtractF",
            y=y,
            df=self.df,
            y_unit=self.y_unit
        )
    
    def integral(self, order: int=1):
        a = self.x * 1j * 2 * np.pi
        b = np.zeros(self.lines, dtype="complex")
        b[1:] = a[1:]**(-order)
        y = self.y * b

        return FreqDomainData(name=f"{self.name}-IntF", y=y, df=self.df, y_unit=self.y_unit+f"*{'s'*order}")    
    
    def effective_value(self, fmin=0, fmax=0):
        # index = (freq > fmin) & (freq <= fmax)
        # effvalue = sqrt(sum(abs(value(index)*new_factor/orig_factor).^2));

        return np.sqrt(np.sum(np.abs(self.y)**2))
    
    def to_timedomain(self):
        single_spec = self.y
        double_spec = np.concatenate([single_spec, np.conjugate(single_spec[self.lines:0:-1])]) / 2
        double_spec[0] *= 2
        # I really need to consider saving all spectrum without converting between ss and ds
        y = np.real(np.fft.ifft(double_spec * len(double_spec)))

        return TimeData(name=self.name, y=y, dt=1/(self.lines*self.df*2), y_unit=self.y_unit)
    
    def as_timedomain(self):
        ...

    def get_amplitudes_at(self, frequencies: list[float], lines: int=3, width: float=None) -> list[tuple[float, float]]:
        if width is not None:
            lines = int(np.ceil(width / self.df))

        x = self.x
        y = self.y
        fas = []

        for f in frequencies:
            i = np.searchsorted(x, f)
            y_p = y[max((i-lines), 0):(i+lines)] # i-lines can <0, and (i-lines):(i+lines) return empty
            x_p = x[max((i-lines), 0):(i+lines)]
            if len(y_p)==0:
                fas.append(None)
                continue
            i_p = np.argmax(np.abs(y_p))
            fas.append( (x_p[i_p], y_p[i_p],) )

        return fas


class FreqIntermediateData(DataBase):
    def __init__(self, name: str = None, uuid: str = None, z: np.ndarray=None, df: float=1, z_unit: str="-", ref_bins: DataBins=None) -> None:
        super().__init__(name, uuid)

        self.z = z if z is not None else np.array([]) # batches x window_size
        self.z_unit = z_unit
        self.df = df
        self.ref_bins = ref_bins

    @property
    def x(self):
        return np.arange(self.lines) * self.df
    
    @property
    def f(self):
        return self.x

    def _bl(self):
        if len(shape:=self.z.shape)==0:
            # shape == ()
            batches, lines = 0, 0
        elif len(shape) == 1: # np.array([p1, p2, p3, ...])
            batches, lines = 1, shape[0]
        else:
            batches, lines = shape

        return batches, lines

    @property
    def lines(self):
        _, lines = self._bl()
        return lines
    
    @property
    def batches(self):
        batches, _ = self._bl()
        return batches
    
    def to_powerspectrum(self, average_by: AverageType=AverageType.Energy):
        if average_by==AverageType.Energy:
            y = np.sqrt(np.mean(np.abs(self.z)**2, axis=0))
        elif average_by==AverageType.Linear:
            y = np.mean(np.abs(self.z), axis=0)
        return FreqDomainData(name=self.name, y=y, df=self.df, y_unit=self.z_unit)
    
    def rectify_to(self, x_slice: tuple, y_slice: tuple) -> "FreqIntermediateData":
        ref_bins = self.ref_bins
        ys = ref_bins.y
        idx = np.argsort(ys)
        ys = ys[idx]
        zs = self.z[idx]

        xs = self.x # the frequencies

        x_bins = np.arange(x_slice)
        x_idxes = np.digitize(xs, x_bins)
        y_bins = np.arange(y_bins)
        y_idxes = np.digitize(ys, y_bins)

        # average by energy
        for y in ys:
            for x in xs:
                pass

        return FreqIntermediateData
    
    def extract_orderslice(self, orders: "OrderList", line_tol: int=3) -> "OrderSliceData":
        xs = self.x # [Hz]
        ys = self.ref_bins.y
        zs = self.z # batch x window

        idx = np.argsort(ys)
        ys = ys[idx]
        zs = zs[idx]

        order_slice = OrderSliceData(name="OrderSlice", source=self)

        for order in orders.orders:
            # slice_element
            se_x = [] # frequency, [Hz]
            se_y = [] # ref value, e.g. [rpm]
            se_z = [] # amplitude, e.g. [mm/s]

            for ref_y, f_batch in zip(ys, zs):
                target_x = ref_y * order.value
                target_idx = np.searchsorted(xs, target_x)

                # TODO: avoid f(0)
                # TODO: avoid out-of-range f
                rel_idx = np.argmax(np.abs(f_batch[max(target_idx-line_tol,0):(target_idx+line_tol)]))
                final_a = f_batch[target_idx-line_tol+rel_idx]
                final_f = xs[target_idx-line_tol+rel_idx]

                se_x.append(final_f)
                # if f already in orderslice? if f not ascending?
                se_y.append(ref_y)
                se_z.append(np.abs(final_a))
                # how to average? by energy

            order_slice.slices[order] = SliceData(f=se_x, ref=se_y, amplitude=se_z)

        return order_slice

    def reference_to(self, reference: "FreqIntermediateData"):
        data = np.conj(reference.z) * self.z / np.abs(reference.z)
        # it's actually rotate self with reference angle
        
        # data = np.mean(data, axis=0)
        # # no linear here, 'cause we can do that later
        # # new object . to_powerspectrum(AverageType.Linear)

        # # calc_phrefspectrum2
        # # I don't know the meaning / scene
        # # it's kind of a different average type
        # data = (
        #         np.mean(np.conj(reference.data) * self.data, axis=0) /
        #         np.sqrt(np.mean(np.abs(reference.data)**2, axis=0))
        #     )

        return FreqIntermediateData(z=data)
    
    def cross_spectrum_with(self, reference: "FreqIntermediateData"):
        
        # assert shape equals, and df, and etc.
        cross = np.conj(reference.z) * self.z
        data = np.mean(cross, axis=0)
        coh = np.sqrt(
                np.abs(data) /
                (self.to_powerspectrum().y * reference.to_powerspectrum().y)
            )
        return FreqDomainData(y=data) # what about coh?
    
    def frf(self, reference: "FreqIntermediateData"):
        cross12 = np.conj(reference.z) * self.z
        cross21 = np.conj(self.z) * reference.z

        spectr1 = np.abs(self.z)**2
        spectr2 = np.abs(reference.z)**2

        frfH1 = np.mean(cross12, axis=0) / np.mean(spectr2, axis=0) # XY/X^2 ???
        frfH2 = np.mean(spectr1, axis=0) / np.mean(cross21, axis=0) # Y^2/XY ???
        # need some theory about: XY v.s. YX
        # and why spectr2 as X^2, spectr1 as Y^2

        return (
            FreqDomainData(y=frfH1),
            FreqDomainData(y=frfH2)
        )

OrderInfo = namedtuple("OrderInfo", ['name', 'value', 'disp_value'])
    # name: label, e.g. f_1
    # value: ratio between reference and frequency
    # disp_value: for the case of unit conversion, e.g. 1st order of 1 [rpm] is actually 1/60 [Hz]

class SliceData:
    def __init__(self, f: np.ndarray, ref: np.ndarray, amplitude: np.ndarray):
        self.f = np.array(f)
        self.ref = np.array(ref)
        self.amplitude = np.array(amplitude)

    def get_aligned_f(self):
        idx = np.argsort(self.f)
        return self.f[idx], self.amplitude[idx]
    
    def get_aligned_ref(self):
        idx = np.argsort(self.ref)
        return self.ref[idx], self.amplitude[idx]

class OrderList(DataBase):
    def __init__(self, name: str = None, uuid: str = None, orders: list[OrderInfo]=None) -> None:
        super().__init__(name, uuid)
        self.orders: list[OrderInfo] = orders or []

class OrderSliceData(DataBase):
    def __init__(self, name: str = None, uuid: str = None, source: FreqIntermediateData = None) -> None:
        super().__init__(name, uuid)

        self.slices: dict[OrderInfo, SliceData] = {}
        self.ref_source: FreqIntermediateData = source
    
    def rectify2freqdata(self): # when speed is uneven, remove high bandwidth
        pass


import gc

from PyQt5 import QtCore, QtGui, QtWidgets
from matplotlib.figure import Figure
from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT, FigureCanvasQTAgg
from matplotlib.backend_bases import NavigationToolbar2

class MultiCanvasNavigation(NavigationToolbar2QT):
    def __init__(self, parent, canvas=None, coordinates=True):
        self._dummy_canvas = canvas or FigureCanvasQTAgg()
        super().__init__(self._dummy_canvas, parent, coordinates)

    def switch_canvas(self, from_canvas, to_canvas):
        if from_canvas is None:
            from_canvas = self._dummy_canvas
        if to_canvas is None:
            to_canvas = self._dummy_canvas

        mode = self.mode
        for _id_name in ("_id_press", "_id_release", "_id_drag"):
            from_canvas.mpl_disconnect(getattr(self, _id_name))
        NavigationToolbar2.__init__(self, to_canvas)
        self.mode = mode

class CanvasDock(QtWidgets.QDockWidget):
    sig_activated = QtCore.pyqtSignal()
    sig_closed = QtCore.pyqtSignal()
    
    class EmbedCanvas(FigureCanvasQTAgg):
        def __init__(self, dock: "CanvasDock", figure=None):
            super().__init__(figure)
            self._dock = dock
        def mousePressEvent(self, event):
            self._dock.sig_activated.emit()
            return super().mousePressEvent(event)

    def __init__(self, name, parent: QtWidgets.QMainWindow):
        super().__init__(name, parent)
        self.setAllowedAreas(QtCore.Qt.RightDockWidgetArea)
        parent.addDockWidget(QtCore.Qt.RightDockWidgetArea, self)
        
        self.figure = figure = Figure()
        self.canvas = canvas = CanvasDock.EmbedCanvas(self, figure)
        self.setContentsMargins(0, 0, 0, 0)
        self.setWidget(canvas)

    def closeEvent(self, event: QtGui.QCloseEvent) -> None:
        self.parent().removeDockWidget(self)
        self.figure.clear()
        gc.collect()
        self.sig_closed.emit()
        return super().closeEvent(event)

    def activate_me(self):
        self.setStyleSheet("background-color: 'orange';")

    def deactivate_me(self):
        p = self.palette()
        p.setColor(self.backgroundRole(), self.parent().palette().color(QtGui.QPalette.Background))
        self.setPalette(p)

class MainWindowWithMCN(QtWidgets.QMainWindow):
    def __init__(self, parent) -> None:
        super().__init__(parent)

        self._mcn = MultiCanvasNavigation(self)
        self._current_canvasdock: CanvasDock = None

        self.setDockOptions(self.AllowNestedDocks)
        tb = self.addToolBar("Plot Navigation")
        tb.addWidget(self._mcn)

    def add_canvasdock(self, name):
        cd = CanvasDock(name, self)
        cd.sig_activated.connect(self._canvasdock_activated)
        cd.sig_closed.connect(self._canvasdock_closed)
        return cd

    def _canvasdock_activated(self):
        cd: CanvasDock = self.sender()
        if self._current_canvasdock is cd:
            return
        if self._current_canvasdock is not None:
            self._current_canvasdock.deactivate_me()
            self._mcn.switch_canvas(self._current_canvasdock.canvas, cd.canvas)
        else:
            self._mcn.switch_canvas(None, cd.canvas)
        cd.activate_me()
        self._current_canvasdock = cd
        
    def _canvasdock_closed(self):
        cd: CanvasDock = self.sender()
        if self._current_canvasdock is cd:
            self._mcn.switch_canvas(cd.canvas, None)
            self._current_canvasdock = None

if __name__=="__main__":
    import sys
    import numpy as np

    app = QtWidgets.QApplication(sys.argv)
    win = MainWindowWithMCN(None)

    for i in range(5):
        cd = win.add_canvasdock(f"Canvas {i}")
        fig = cd.figure
        ax = fig.gca()
        ax.plot(np.random.random(10_000))

    win.show()
    app.exit(app.exec())


import re, yaml
from os import path
from dac.core import Container, NodeBase

def use_plugin(setting_fpath: str, clean: bool=True, dac_win=None):
    alias_pattern = re.compile("^/(?P<alias_name>.+)/(?P<rest>.+)")
    def get_node_type(cls_path: str) -> str | type[NodeBase]:
        if cls_path[0]=="[" and cls_path[-1]=="]":
            return cls_path # just str as section string
        
        if (rst:=alias_pattern.search(cls_path)):
            cls_path = alias[rst['alias_name']]+"."+rst['rest']

        try:
            return Container.GetClass(cls_path)
        except AttributeError:
            dac_win.message(f"Module `{cls_path}` not found")
            return None
        
    if clean:
        Container._action_types.clear()
        Container._key_types.clear()
        # quick_tasks and quick_actions are always overwritten

    with open(setting_fpath, mode="r", encoding="utf8") as fp:
        setting: dict = yaml.load(fp, Loader=yaml.FullLoader)
        if not setting: return

        if (inherit_rel_path:=setting.get('inherit')) is not None:
            use_plugin(path.join(path.dirname(setting_fpath), inherit_rel_path), clean=False, dac_win=dac_win)

        alias = setting.get('alias', {})

        for gdts in setting.get('data', {}).get("_", []): # global_data_type_string
            node_type = get_node_type(gdts)
            if node_type: Container.RegisterGlobalDataType(node_type)

        for dts, catss in setting.get('actions', {}).items(): #  data_type_string, context_action_type_string_s
            if dts=="_": # global_context
                for cats in catss:
                    node_type = get_node_type(cats)
                    if node_type: Container.RegisterGlobalContextAction(node_type)
            else:
                data_type = get_node_type(dts)
                if not data_type: continue
                for cats in catss:
                    action_type = get_node_type(cats)
                    if action_type: Container.RegisterContextAction(data_type, action_type)

        if not hasattr(dac_win, "show"): # web-based cannot use PyQt5 and the tasks
            return

        for ats, tss in setting.get("quick_tasks", {}).items(): # action_type_string, task_string_s
            action_type = get_node_type(ats)
            if not action_type: continue
            action_type.QUICK_TASKS = [] # make superclass.QUICK_TASKS hidden
            for tts, name, *rest in tss: # task_type_string, name, *rest
                task_type = get_node_type(tts)
                if not task_type: continue
                task = task_type(dac_win=dac_win, name=name, *rest)
                action_type.QUICK_TASKS.append(task)

        for ats, (tts, name, *rest) in setting.get("default_task", {}).items(): # action_type_string, task_type_string
            action_type = get_node_type(ats)
            if not action_type: continue
            action_type.DEFAULT_TASK = None
            task_type = get_node_type(tts)
            if not task_type: continue
            task = task_type(dac_win=dac_win, name=name, *rest)
            action_type.DEFAULT_TASK = task

        for dts, ass in setting.get("quick_actions", {}).items(): # data_type_string, action_string_s
            data_type = get_node_type(dts)
            if not data_type: continue
            data_type.QUICK_ACTIONS = []
            for ats, dpn, opd in ass: # action_type_string, data_param_name, other_params_dict
                action_type = get_node_type(ats)
                if not action_type: continue
                data_type.QUICK_ACTIONS.append((action_type, dpn, opd))


import numpy as np
from . import BallBearing, GearboxDefinition, BearingInputStage
from dac.modules.timedata import TimeData
from dac.core.actions import VAB, PAB, SAB, ActionBase
from dac.modules.timedata.actions import ShowTimeDataAction
from dac.modules.nvh.data import OrderList, OrderInfo
from dac.modules.nvh.actions import ViewFreqDomainAction
from matplotlib.backend_bases import MouseButton, MouseEvent

class CreateBearing(ActionBase):
    CAPTION = "Make a bearing"
    def __call__(self, N_balls: int=8, D_ball: float=2, D_pitch: float=12, beta: float=15) -> BallBearing:
        return BallBearing(
            name="Ball bearing",
            N_balls=N_balls,
            D_ball=D_ball,
            D_pitch=D_pitch,
            beta=beta,
        )

class CreateGearboxWithBearings(ActionBase):
    CAPTION = "Make gearbox with bearings"
    def __call__(self, gearbox: GearboxDefinition, bearings: list[tuple[BearingInputStage, BallBearing]]) -> GearboxDefinition:
        return GearboxDefinition(
            name="Gearbox with bearings",
            stages=gearbox.stages.copy(),
            bearings=bearings,
        )


class CreateOrdersOfGearbox(ActionBase):
    CAPTION = "Create orders for gearbox"
    def __call__(self, gearbox: GearboxDefinition, ref_output: bool=True) -> OrderList:
        ol = OrderList(f"Orders-{gearbox.name}")
        for freq, label in gearbox.get_freqs_labels_at(speed=60, speed_on_output=ref_output):
            # reference shaft has order 1
            ol.orders.append(OrderInfo(label, freq/60, freq))

        return ol

class ShowFreqLinesTime(VAB):
    CAPTION = "Mark frequency lines on time domain"
    def __call__(self, gearbox: GearboxDefinition, speed_channel: TimeData, speed_on_output: bool=True, stages: list[int]=[1, 2], fmt_lines: list[str]=["{f_1}", "{f_2}-{f_1}"]): # bearings: list[tuple[BallBearing, BearingInputStage]]
        if not speed_channel or not gearbox:
            return
        
        if stages is None:
            stages = []
        if fmt_lines is None:
            fmt_lines = []

        canvas = self.canvas
        widgets = [] # it's actually patches

        def on_press(event: MouseEvent):
            if ( (not (ax:=event.inaxes)) or event.button!=MouseButton.LEFT or canvas.widgetlock.locked() ):
                return
            for widget in widgets: # widgets from previous press
                widget.remove()
            widgets.clear()

            bits = 0
            for stage_num in stages:
                bits |= 1<<(stage_num-1)
            moment = event.xdata

            trans = ax.get_xaxis_text1_transform(0)
            speed = np.abs(np.mean(speed_channel.y)) # if isnumber(speed_channel), just assign

            for freq, label in gearbox.get_freqs_labels_at(speed, speed_on_output, choice_bits=bits):
                dt = 1 / freq
                x = moment + dt

                widgets.append( ax.axvline(x, ls="--", lw=1) )
                widgets.append( ax.text(x, 1, label, transform=trans[0]) )

            format_dict = {label: freq for freq, label in gearbox.get_freqs_labels_at(speed, speed_on_output)}
            for i, fmt_line in enumerate(fmt_lines):
                label, *freqs = fmt_line.split(",", maxsplit=1)

                if freqs: # freq provided
                    freq = float(freqs[0])
                else:
                    freq = eval(label.format(**format_dict))

                dt = 1 / freq
                x = moment + dt
                                
                widgets.append( ax.axvline(x, ymax=0.95-0.05*(i%2), ls="--", lw=1) )
                widgets.append( ax.text(x, 0.95-0.05*(i%2), label, transform=trans[0]) )

            widgets.append(ax.axvline(event.xdata))
            canvas.draw_idle()

        self._cids.append( canvas.mpl_connect("button_press_event", on_press) )

class ShowFreqLinesFreq(VAB):
    CAPTION = "Mark frequency lines on spectrum"
    def __call__(self, gearbox: GearboxDefinition, speed_channel: TimeData, speed_on_output: bool=True, stages: list[int]=[1, 2], fmt_lines: list[str]=["{f_1}", "{f_2}-{f_1}"]):
        if not speed_channel or not gearbox:
            return
        
        if stages is None:
            stages = []
        if fmt_lines is None:
            # `fmt_lines`, e.g.
            # {f_2}-{f_1}
            # f_custom, 1.1
            # TODO: f_custom, {f_2}-{f_1} # and usable without `speed_channel` or `gearbox`

            fmt_lines = []
        
        fig = self.figure
        ax = fig.gca()

        canvas = self.canvas
        widgets = [] # it's actually patches

        def plot_lines(start_freq: float, sideband: bool=False):
            for widget in widgets: # widgets from previous press
                widget.remove()
            widgets.clear()

            bits = 0
            for stage_num in stages:
                bits |= 1<<(stage_num-1)

            trans = ax.get_xaxis_text1_transform(0)
            speed = np.abs(np.mean(speed_channel.y)) # if isnumber(speed_channel), just assign

            delta_factors = [1] if not sideband else [1, -1]

            for freq, label in gearbox.get_freqs_labels_at(speed, speed_on_output, choice_bits=bits):
                # TODO: based on checkbox
                # if 'fz' in label:
                #     continue
                for factor in delta_factors:
                    widgets.append(ax.axvline(start_freq+freq*factor, ls="--", lw=1))
                    widgets.append(ax.text(start_freq+freq*factor, 1, label, transform=trans[0]))

            format_dict = {label: freq for freq, label in gearbox.get_freqs_labels_at(speed, speed_on_output)}
            for i, fmt_line in enumerate(fmt_lines):
                label, *freqs = fmt_line.split(",", maxsplit=1)

                if freqs: # freq provided
                    freq = float(freqs[0])
                else:
                    freq = eval(label.format(**format_dict))

                for factor in delta_factors:
                    widgets.append(ax.axvline(start_freq+freq*factor, ymax=0.95-0.05*(i%2), ls="--", lw=1))
                    widgets.append(ax.text(start_freq+freq*factor, 0.95-0.05*(i%2), label, transform=trans[0]))
            
            if sideband:
                widgets.append(ax.axvline(start_freq))
            canvas.draw_idle()

        def on_press(event: MouseEvent):
            if ( (not (ax:=event.inaxes)) or canvas.widgetlock.locked() ):
                return
            if event.button==MouseButton.LEFT:
                plot_lines(event.xdata, sideband=True)
            elif event.button==MouseButton.RIGHT:
                plot_lines(0, sideband=False)

        plot_lines(0, sideband=False) # init plot

        self._cids.append( canvas.mpl_connect("button_press_event", on_press) )

class ShowSpectrumWithFreqLines(SAB, seq=[ViewFreqDomainAction, ShowFreqLinesFreq]):
    CAPTION = "Show FFT spectrum with freq lines"

class ShowTimeDataWithFreqLines(SAB, seq=[ShowTimeDataAction, ShowFreqLinesTime]):
    CAPTION = "Show time data with freq lines"


import numpy as np

def to_decibel(value, ref, by_amplitude=True):
    if by_amplitude: # ref = 20 uPa
        return 20*np.log10(value / ref)
    else: # by_power # ref = 10^-12 W
        return 10*np.log10(value / ref)

def a_weighted(freq: np.array | float) -> np.array | float:
    # dB + diff => dB(A)
    
    A_diff = 20 * np.log10(
        (12194**2 * freq**4) /
        ( (freq**2 + 20.6**2) * (((freq**2 + 107.7**2) * (freq**2 + 737.9**2))**0.5) * (freq**2 + 12194**2) )
    ) + 2
    # 20*log_10(RA(f)) - 20*log_10(RA(1000)) = 20*log_10(RA(f)) - (-2)

    return A_diff


from dac.core import DataNode, ContextKeyNode

r"""Concepts and principles

- Workable without container
- Data types are directly use-able as action required
- Member types are direct objects (not the identifier to get target object from container)
- "Definition"s contain just basic elements (int/float/str/dict/list/bool/...)
- Provide hint (<type hint> or [literal hint]) for members
"""

class DataBase(DataNode):
    QUICK_ACTIONS = []
    # the actions to be performed on individual data node
    # equal to `ActionBase(win, fig, ...)(DataBase())`

class SimpleDefinition(ContextKeyNode):
    pass

class ReferenceBase:
    pass

class StatData:
    pass


import traceback, sys, inspect
from PyQt5.QtCore import QRunnable, QObject, pyqtSignal, pyqtSlot, QMutex

# https://gist.github.com/ksvbka/1f26ada0c6201c2cf19f59b100d224a9

class WorkerSignals(QObject):
    '''
    Defines the signals available from a running worker thread.
    Supported signals are:
    - started: No data
    - finished: No data
    - error: `exctype, value, traceback`
    - result: `object` data returned from processing, anything
    - progress: `int, int` indicating progress metadata
    - message: `str` the message from worker
    '''

    started = pyqtSignal()
    finished = pyqtSignal()
    error = pyqtSignal(type, object, object)
    result = pyqtSignal(object)
    progress = pyqtSignal(int, int)
    message = pyqtSignal(str)

class ThreadWorker(QRunnable):
    '''
    Worker thread
    Inherits from QRunnable to handler worker thread setup, signals and wrap-up.
    '''

    def __init__(self, fn, caption: str, *args, mutex: QMutex=None, **kwargs):
        super().__init__()
        self.caption = caption
        self.mutex = mutex

        # Store constructor arguments (re-used for processing)
        self.fn = fn
        self.args = args
        self.kwargs = kwargs

        self.signals = WorkerSignals()

    @pyqtSlot()
    def run(self):
        '''
        Initialise the runner function with passed args, kwargs.
        '''
        # Retrieve args/kwargs here; and fire processing using them
        
        # Add the callback to our kwargs
        fn_params = inspect.signature(self.fn).parameters
        if 'progress_emitter' in fn_params:
            self.kwargs['progress_emitter'] = self.signals.progress.emit
        if 'logger' in fn_params:
            self.kwargs['logger'] = self.signals.message.emit

        try:
            if self.mutex is not None:
                self.mutex.lock()
            self.signals.started.emit()
            
            result = self.fn(*self.args, **self.kwargs)
        except:
            # traceback.print_exc()
            # exctype, value, tb = sys.exc_info()
            self.signals.error.emit(*sys.exc_info())
        else:
            self.signals.result.emit(result)  # Return the result of the processing
        finally:
            if self.mutex is not None:
                self.mutex.unlock()
            self.signals.finished.emit()  # Done


from nptdms import TdmsFile, TdmsGroup, TdmsChannel
from nptdms import TdmsWriter, RootObject, GroupObject, ChannelObject

from collections import OrderedDict
from . import TimeData
from datetime import datetime, timedelta, timezone

def load_tdms(fpath) -> list[TimeData]:
    r = []
    f = TdmsFile(fpath, read_metadata_only=False, keep_open=False)
    for g in f.groups():
        g: TdmsGroup
        for c in g.channels():
            c: TdmsChannel
            prop: OrderedDict = c.properties

            gain = float(prop['Gain'])
            offset = float(prop['Offset'])
            y_unit = prop['Unit']
            desc = prop['Description']
            x_unit = prop['wf_xunit_string']
            dt = float(prop['wf_increment'])
            # length = prop['wf_samples']

            r.append(TimeData(name=c.name, y=c.data, dt=dt, y_unit=y_unit, comment=desc))

    return r

def save_tdms(channels: list[TimeData], fpath: str, start_time: datetime, group_name: str="group_1"):
    current_tz = datetime.now().astimezone().tzinfo
    start_time = start_time.replace(tzinfo=current_tz)

    with TdmsWriter(fpath) as tdms_writer:
        root_object = RootObject()
        group_object = GroupObject(group_name)

        for channel in channels:
            channel_object = ChannelObject(group_name, channel.name, channel.y, properties={
                "unit_string": channel.y_unit,
                "wf_xunit_string": "s",
                "wf_xname": "Time",
                "wf_start_time": start_time,
                "wf_increment": channel.dt,
                "wf_samples": channel.length
            })

            tdms_writer.write_segment([
                root_object,
                group_object,
                channel_object,
            ])


from os import path
from PyQt5 import QtWidgets, QtCore
from dac.core.actions import ActionBase
from dac.gui import TaskBase
from dac import APPNAME

APPSETTING = QtCore.QSettings(APPNAME, "TimeData")
SET_RECENTDIR = "RecentDir"

class FillFpathsTask(TaskBase):
    def __call__(self, action: ActionBase):
        fpaths, fext = QtWidgets.QFileDialog.getOpenFileNames(
            self.dac_win, caption="Select measurement files",
            directory=APPSETTING.value(SET_RECENTDIR)
        )
        if not fpaths:
            return
        APPSETTING.setValue(SET_RECENTDIR, path.dirname(fpaths[0]))

        action._construct_config['fpaths'] = fpaths


import numpy as np
from collections import namedtuple

from dac.core.actions import ActionBase
from . import TimeData

CosineComponent = namedtuple("CosineComponent", ['freq', 'amp', 'phase'])

class SignalConstructAction(ActionBase):
    CAPTION = "Construct signal with cosines"
    def __call__(self, components: list[CosineComponent], offset: float=0, duration: float=10, fs: int=1000) -> TimeData:
        r"""Construct time domain data with cosine waves.

        Parameters
        ----------
        components : [(frequency, amplitude, phase)]
            list of tuples, each tuple contains basic info of the cosine wave

            frequency: float, [Hz]

            amplitude: float

            phase: float, [°]

        fs : int, [Hz]
            sample rate

        duration : float, [s]
            sample time
            
        """

        t = np.arange(int(duration * fs)) / fs
        y = np.zeros_like(t) + offset
        
        for freq, amp, phase in components:
            y += amp*np.cos(2*np.pi*freq*t + np.deg2rad(phase))

        return TimeData(name="Generated signal", y=y, dt=1/fs, y_unit="-", comment="Constructed time data")


from enum import Enum
import numpy as np
from dac.core.data import DataBase
from collections import defaultdict

class GearStage:
    class StageType(Enum):
        Unknown = 0
        Planetary = 1
        Parallel = 2
    
    def __init__(self, config: dict):
        self.config = config
        match config:
            case {"RG": rg, "PG": pg, "SU": su, "NoP": nop}:
                ratio = 1 + rg/su
                t = GearStage.StageType.Planetary
            case {"Wheel": wheel, "Pinion": pinion}:
                ratio = wheel / pinion
                t = GearStage.StageType.Parallel
            case _:
                ratio = 1
                t = GearStage.StageType.Unknown

        self.ratio = ratio
        self.stage_type = t

    def f(self, input_speed): # order=1
        return input_speed / 60

    def fz(self, input_speed, order=1):
        if self.stage_type==GearStage.StageType.Planetary:
            n = self.config['NoP']
            z = self.config['RG']*order
            Z = round(z/n) * n # in non-factorizing case, find closest (dominant) side band; for factorizing, same as `z`
            _Z = (-1)**(Z>z)*n*(Z!=z)+Z # the other side band
            # _Z = np.sign(z-Z)*n+Z # same

            # TODO: distinguish fz and fzN/fzn, otherwise 2*{fz} is not {2fz}
            
            return input_speed / 60 * Z
        
        elif self.stage_type==GearStage.StageType.Parallel:
            return input_speed / 60 * self.config['Wheel']

    def input_speed(self, output_speed):
        return output_speed / self.ratio

    def get_freq_at_order(self, order: str):
        ...

    def get_order_name_at_freq(self, freq: float, max_order: int, tolerance: int):
        # return the name
        ...

    def get_freqs_labels_at(self, speed: float, speed_on_output: bool=True):
        rst = [
            (
                self.f( self.input_speed(speed) if speed_on_output else speed ),
                "f"
            ),
            (
                self.fz( self.input_speed(speed) if speed_on_output else speed ),
                "fz"
            ),
        ]
        if self.stage_type==GearStage.StageType.Planetary:
            NoP = self.config['NoP']
            rst.append(
                (
                    NoP * self.f( self.input_speed(speed) if speed_on_output else speed ),
                    f"{NoP}pf",
                )
            )
            rst.append(
                (
                    self.f( self.input_speed(speed) if speed_on_output else speed ) * self.config['RG'] / self.config['PG'], # PG rotation speed related to PC
                    f"fw",
                )
            )

        return rst

class GearboxDefinition(DataBase):
    def __init__(self, name: str = None, uuid: str = None, stages: list[GearStage]=None, bearings: list[tuple["BearingInputStage", "BallBearing"]]=None) -> None:
        super().__init__(name, uuid)

        self.stages = stages or []
        self.bearings = bearings or []

    def get_construct_config(self) -> dict:
        if self.stages:
            stgs = [stg.config for stg in self.stages]
        else:
            stgs = [
                {"RG": "<int>", "PG": "<int>", "SU": "<int>", "NoP": "[num of planets]"},
                {"Wheel": "<int>", "Pinion": "<int>"},
            ]

        return {
            "name": self.name,
            "stages": stgs
        }
    
    def apply_construct_config(self, construct_config: dict):
        self.name = construct_config["name"]
        self.stages.clear()
        for stage in construct_config["stages"]:
            self.stages.append(GearStage(stage))

    def get_freqs_labels_at(self, speed: float, speed_on_output: bool=True, choice_bits: int=-1):
        rst = []
        bearing_dict = defaultdict(list[BallBearing])
        for stage, bearing in self.bearings:
            bearing_dict[stage].append(bearing)

        if not speed_on_output:
            for i, stg in enumerate(self.stages):
                if (choice_bits==-1) or (choice_bits & (1<<i)):
                    for freq, lbl in stg.get_freqs_labels_at(speed, speed_on_output):
                        rst.append(
                            (freq, f"{lbl}_{i+1}",)
                        )

                    for bearing in bearing_dict[i+1]:
                        for freq, lbl in bearing.get_freqs_labels_at(speed):
                            rst.append(
                                (freq, f"{lbl}_{bearing.name}",)
                            )
                speed = speed * stg.ratio
        else:
            n = len(self.stages)
            for i, stg in enumerate(self.stages[::-1]):
                if (choice_bits==-1) or (choice_bits & (1<<(n-i-1))):
                    for freq, lbl in stg.get_freqs_labels_at(speed, speed_on_output):
                        rst.append(
                            (freq, f"{lbl}_{n-i}",)
                        )
                speed = speed / stg.ratio

                # always bearing on input shaft
                # need to make a virtual 1:1 stage for final output
                if (choice_bits==-1) or (choice_bits & (1<<(n-i-1))):
                    for bearing in bearing_dict[n-i]:
                        for freq, lbl in bearing.get_freqs_labels_at(speed):
                            rst.append(
                                (freq, f"{lbl}_{bearing.name}",)
                            )

        # TODO: force regarding input shaft to simplify code

        return rst

class BearingInputStage(int): # cannot use namedtuple(BallBearing, BearingInputStage) because no conversion for namedtuple
    # always the input speed stage
    pass

class BallBearing(DataBase):
    def __init__(self, name: str = None, uuid: str = None, N_balls: int=8, D_ball: float=2, D_pitch: float=12, beta: float=15) -> None:
        super().__init__(name, uuid)
        self.N_balls = N_balls
        self.D_ball = D_ball
        self.D_pitch = D_pitch # = (D_IR+D_OR)/2
        self.beta = beta

    def get_freqs_labels_at(self, speed: float):
        freq = speed / 60
        return [
            (self.bpfo()*freq, 'bpfo',),
            (self.bpfi()*freq, 'bpfi',),
            (self.bsf()*freq, 'bsf',),
            (self.ftf()*freq, 'ftf',),
        ]

    def bpfo(self):
        # outer race defect frequency
        # ~ 0.4 * N_balls
        return self.N_balls/2 * (1-self.D_ball/self.D_pitch*np.cos(np.deg2rad(self.beta)))
    
    def bpfi(self):
        # inner race defect frequency
        # ~ 0.6 * N_balls
        return self.N_balls/2 * (1+self.D_ball/self.D_pitch*np.cos(np.deg2rad(self.beta)))
    
    def bsf(self):
        # ball defect frequency
        return self.D_pitch/self.D_ball*(1-(self.D_ball/self.D_pitch*np.cos(np.deg2rad(self.beta)))**2)
    
    def ftf(self):
        # cage defect frequency
        # ~ 0.4
        return 1/2*(1-self.D_ball/self.D_pitch*np.cos(np.deg2rad(self.beta)))


from enum import Enum
import numpy as np

class WindowType(Enum):
    Uniform = (1.0, 1.0)
    Hanning = (2.0, 1.63)
    Hamming = (1.85, 1.59)
    KaiserBessel = (2.49, 1.86)
    FlatTop = (4.18, 2.26)
    Blackman = (2.80, 1.97)
    Unknown = (1.0, 1.0) # (BandCorrection.NarrowBand, BandCorrection.BroadBand)

class BandCorrection(Enum):
    NarrowBand = 0 # amplitude correction
    BroadBand = 1 # power correction

class AmplitudeType(Enum):
    RMS = 1/np.sqrt(2)
    Peak = 1

class AverageType(Enum):
    Energy = 0
    Linear = 1
    EnergyExponential = 2
    Maximum = 3
    Minimum = 4

class FilterType(Enum):
    LowPass = "lowpass"
    HighPass = "highpass"
    BandPass = "bandpass"
    BandStop = "bandstop"

class BinMethod(Enum):
    Mean = 0
    Min = 1

class ToleranceType(Enum):
    FixLines = 0
    FixBandWidth = 1
    FixOrderWidth = 2
    FixFrequencyPercentage = 3


import numpy as np
from collections import defaultdict
from matplotlib.gridspec import GridSpec
from matplotlib.widgets import TextBox, RadioButtons
from matplotlib.axes import Axes

from dac.core.data import SimpleDefinition
from dac.core.actions import ActionBase, VAB, PAB, SAB
from dac.modules.timedata import TimeData
from . import WindowType, BandCorrection, BinMethod, AverageType, ToleranceType
from .data import FreqIntermediateData, DataBins, FreqDomainData, \
                  OrderInfo, OrderList, SliceData, OrderSliceData

class ToFreqDomainAction(PAB):
    CAPTION = "Simple FFT to frequency domain" # rect window

    def __call__(self, channels: list[TimeData], window: WindowType=WindowType.Uniform, corr: BandCorrection=BandCorrection.NarrowBand) -> list[FreqDomainData]:
        rst = []
        m = len(channels)
        window_funcs = {
            WindowType.Hanning: np.hanning,
            WindowType.Hamming: np.hamming,
        }
        for i, ch in enumerate(channels):
            batch_N = ch.length
            df = 1 / (batch_N*ch.dt)
            
            if window in window_funcs:
                windowed_y = ch.y * window_funcs[window](batch_N)
            else:
                window = WindowType.Uniform
                windowed_y = ch.y
            fdata = np.fft.fft(windowed_y) / batch_N * window.value[corr.value]

            double_spec = fdata[:int(np.ceil(batch_N/2))]
            double_spec[1:] *= 2

            rst.append(FreqDomainData(name=ch.name, y=double_spec, df=df, y_unit=ch.y_unit))
            self.progress(i+1, m)
        return rst

class ToFreqIntermediateAction(PAB):
    CAPTION = "FFT to frequency domain with window and reference"

    def __call__(self, channels: list[TimeData],
                 window: WindowType=WindowType.Hanning, corr: BandCorrection=BandCorrection.NarrowBand,
                 resolution: float=0.5, overlap: float=0.75,
                 ref_channel: TimeData=None,
                ) -> list[FreqIntermediateData]:
        
        freqs = []

        window_funcs = {
            WindowType.Hanning: np.hanning,
            WindowType.Hamming: np.hamming,
        }

        if ref_channel is not None:
            ref_batches = ref_channel.to_bins(df=resolution, overlap=overlap)
            ref_bins_y = np.mean(ref_batches, axis=1)
            ref_bins = DataBins(name=ref_channel.name, y=ref_bins_y, y_unit=ref_channel.y_unit)
        # else:
        #     create a TimeData channel, but don't know the length

        n = len(channels)
        for i, channel in enumerate(channels):
            batches = channel.to_bins(df=resolution, overlap=overlap)
            N_batches, batch_N = batches.shape

            if ref_channel is None:
                ref_bins_y = np.arange(N_batches) * 1/resolution * (1-overlap)
                ref_bins = DataBins(name="Time", y=ref_bins_y, y_unit="s")
                ref_bins._method = BinMethod.Min

            batches = batches * window_funcs[window](batch_N)
            batches_fft = np.fft.fft(batches) / batch_N * window.value[corr.value]

            double_spec = batches_fft[:, :int(np.ceil(batch_N/2))]
            double_spec[:, 1:] *= 2

            freq = FreqIntermediateData(name=channel.name, z=double_spec, df=resolution, z_unit=channel.y_unit, ref_bins=ref_bins)
            freqs.append(freq)
            self.progress(i+1, n)

        return freqs

class AverageIntermediateAction(ActionBase):
    CAPTION = "Average (static) FreqIntermediate to spectrum"
    def __call__(self, channels: list[FreqIntermediateData], average_by: AverageType=AverageType.Energy) -> list[FreqDomainData]:
        rst = []
        for channel in channels:
            rst.append(channel.to_powerspectrum(average_by=average_by))
        return rst
    
class ViewFreqDomainAction(VAB):
    CAPTION = "Show FFT spectrum"

    def __call__(self, channels: list[FreqDomainData], xlim: tuple[float, float]=None, ylim: tuple[float, float]=None, with_phase: bool=False):
        fig = self.figure
        gs = GridSpec(2, 1, height_ratios=[2, 1])

        if with_phase:
            ax = fig.add_subplot(gs[0])
            ax_p = fig.add_subplot(gs[1], sharex=ax)
            ax_p.set_ylabel("Phase [°]")
        else:
            ax = fig.gca()

        ax.set_xlabel("Frequency [Hz]")
        ax.set_ylabel("Amplitude")

        if xlim: ax.set_xlim(xlim)
        if ylim: ax.set_ylim(ylim)

        for channel in channels:
            ax.plot(channel.x, channel.amplitude, label=f"{channel.name} [{channel.y_unit}]")
            if with_phase:
                ax_p.plot(channel.x, channel.phase)

        ax.legend(loc="upper right")

class ViewFreqIntermediateAction(VAB):
    CAPTION = "Show FFT color plot"

    def __call__(self, channel: FreqIntermediateData, xlim: tuple[float, float]=None, clim: tuple[float, float]=[0, 0.001]):
        fig = self.figure
        ax = fig.gca()

        if clim is None:
            clim = [None, None]
        cmin, cmax = clim

        fig.suptitle(f"Color map: {channel.name}")
        xs = channel.x
        zs = channel.z
        ax.set_xlabel("Frequency [Hz]")
        if (ref_bins:=channel.ref_bins) is not None:
            ys = channel.ref_bins.y
            idx = np.argsort(ys)
            ys = ys[idx]
            zs = zs[idx]
            ax.set_ylabel(f"{ref_bins.name} [{ref_bins.y_unit}]")
        m = ax.pcolormesh(xs, ys, np.abs(zs), cmap='jet', vmin=cmin, vmax=cmax)
        cb = fig.colorbar(m)
        cb.set_label(f"Amplitude [{channel.z_unit}]")
        if xlim is not None:
            ax.set_xlim(xlim)

class ExtractAmplitudeAction(PAB):
    CAPTION = "Extract amplitude at frequencies"

    def __call__(self, channels: list[FreqDomainData], frequencies: list[float], line_tol: int=3):
        ...

class ViewColorPlotAndCheckOrderSlice(ViewFreqIntermediateAction):
    # obsolete, it's too laggy
    CAPTION = "Show color plot with order indication"
    def __call__(self, channel: FreqIntermediateData, xlim: tuple[float, float] = None, clim: tuple[float, float] = [0, 0.001]):
        super().__call__(channel, xlim, clim)
        fig = self.figure
        ax: Axes = fig.axes[0]
        
        ax_order = fig.add_axes([0.1, 0.01, 0.5, 0.05])
        order_input = TextBox(ax_order, label="Order:", initial=0)

        l = ax.axline((0, 0), slope=0, color='k')

        def on_press(event):
            if fig.canvas.widgetlock.locked():
                return            
            order_input.set_val(f"{event.ydata/event.xdata:.3f}")

        def on_order_submit(order_str: str):
            order = float(order_str)
            if order<=0:
                return
            l.set_slope(order)
            fig.canvas.draw_idle()
            
        order_input.on_submit(on_order_submit)

        self._cids.append(
            fig.canvas.mpl_connect('button_press_event', on_press)
        )
        self._widgets.append(order_input)

class MarkOrders(VAB):
    pass

class ViewColorPlotWithOrderSlice(ViewFreqIntermediateAction): # inherit SAB = ViewColorPlot + MarkOrders
    CAPTION = "Show color plot with order indication"
    def __call__(self, channel: FreqIntermediateData, orders: OrderList, fmt_lines: list[str]=["{f_1}", "0.5"], xlim: tuple[float, float] = None, clim: tuple[float, float] = [0, 0.001]):
        pass

class CreateOrders(ActionBase):
    CAPTION = "Create orders"

    def __call__(self, infos: list[OrderInfo]) -> OrderList:
        ol = OrderList(name="Orders")

        for name, value, disp_value in infos:
            if disp_value is None:
                disp_value = value
            ol.orders.append(OrderInfo(name, value, disp_value))

        return ol

class ExtractOrderSlicesAction(PAB):
    CAPTION = "Extract OrderSlice"

    def __call__(self, channels: list[FreqIntermediateData], orders: OrderList, tol_type: ToleranceType=ToleranceType.FixLines, tol_value: float=3) -> list[OrderSliceData]:
        order_slices = []
        for channel in channels:
            os = channel.extract_orderslice(orders, int(tol_value))
            os.name = f"OrderSlice-{channel.name}"
            order_slices.append(os)

        return order_slices

# pick up order from color plot

class ViewOrderSlice(VAB):
    CAPTION = "View OrderSlice"
    def __call__(self, order_slice: OrderSliceData):
        # switch orders
        # by reference / by frequency
        fig = self.figure
        ax = fig.gca()

        ax_orders = fig.add_axes([0.91, 0.13, 0.08, 0.2])
        ax_method = fig.add_axes([0.91, 0.01, 0.08, 0.1])
        order_labels = ["None"] + [order.name for order in order_slice.slices]
        order_choice: OrderInfo | None = None
        order_selector = RadioButtons(ax_orders, order_labels, active=0)
        method_labels = ['By frequency', 'By reference']
        method_choice = 0
        method_selector = RadioButtons(ax_method, method_labels, active=method_choice)

        def order_change(label):
            nonlocal order_choice
            for order in order_slice.slices:
                if order.name==label:
                    order_choice = order
                    break
            else:
                order_choice = None
            update()
        order_selector.on_clicked(order_change)
        def method_change(label):
            nonlocal method_choice
            method_choice = method_labels.index(label)
            update()
        method_selector.on_clicked(method_change)

        def update():
            ax.cla()
            if order_choice is None:
                fig.canvas.draw_idle()
                return
            slice = order_slice.slices[order_choice]
            if method_choice==0: # by freq
                x, y = slice.get_aligned_f()
            else:
                x, y = slice.get_aligned_ref()
            ax.plot(x, y)

            fig.canvas.draw_idle()

        self._widgets.append(method_selector)
        self._widgets.append(order_selector)

class ViewOrderSliceOfMeasurements(VAB):
    CAPTION = "View OrderSlice of measurements"
    def __call__(self, measurements: list[SimpleDefinition], orderslice_name: str):
        fig = self.figure
        ax = fig.gca()
        ax.set_title(orderslice_name)

        fiad = defaultdict(list) # fork_in_another_direction -_-,

        for measurement in measurements:
            ctx = self.container.get_context(measurement)
            orderslice: OrderSliceData = ctx.get_node_of_type(orderslice_name, OrderSliceData)
            
            for oi, sd in orderslice.slices.items():
                fiad[oi.name].append( (measurement.name, sd,) )

        ax_orders = fig.add_axes([0.91, 0.13, 0.08, 0.2])
        order_names = ["None"] + list(fiad.keys())
        order_choice:str = ""
        order_selector = RadioButtons(ax_orders, order_names, active=0)
        ax_method = fig.add_axes([0.91, 0.01, 0.08, 0.1])
        method_labels = ['By frequency', 'By reference']
        method_choice:int = 0
        method_selector = RadioButtons(ax_method, method_labels, active=method_choice)
        def order_change(label):
            nonlocal order_choice
            order_choice = label
            update()
        order_selector.on_clicked(order_change)
        def method_change(label):
            nonlocal method_choice
            method_choice = method_labels.index(label)
            update()
        method_selector.on_clicked(method_change)

        def update():
            ax.cla()
            ms = fiad.get(order_choice)
            if not ms:
                fig.canvas.draw_idle()
                return
            for (measurement_name, slice_data,) in ms:
                slice_data: SliceData
                if method_choice==0: # by freq
                    x, y = slice_data.get_aligned_f()
                else:
                    x, y = slice_data.get_aligned_ref()
                ax.plot(x, y, label=f"{measurement_name}")
            ax.legend(loc='upper right')
            fig.canvas.draw_idle()

        self._widgets.append(order_selector)
        self._widgets.append(method_selector)

# calc rms

class FilterSpectrumAction(PAB):
    CAPTION = "Filter spectrum"
    def __call__(self, channels: list[FreqDomainData], bands: list[tuple[float, float]], remove: bool=True) -> list[FreqDomainData]:
        rst = []
        if remove:
            for ch in channels:
                rst.append(ch.remove_spec(bands=bands))
        else:
            for ch in channels:
                rst.append(ch.keep_spec(bands=bands))
        return rst
    
class SpectrumToTimeAction(PAB):
    CAPTION = "Convert spectrum to TimeData"
    def __call__(self, channels: list[FreqDomainData]) -> list[TimeData]:
        rst = []
        for ch in channels:
            rst.append(ch.to_timedomain())
        return rst

class SpectrumAsTimeAction(PAB):
    CAPTION = "Treate frequency spectrum as TimeData"

class LoadCaseSpectrumComparison(VAB):
    def __call__(self, loadcases: list[str], channel_name: str):
        pass

class LoadCaseFreqIntermediateAverage(VAB):
    def __call__(self, loadcases: list[str], channel_name: str, ref_case: str):
        pass

# BearingEnvelopeAnalysis = FFT + FilterSpec + ...


import inspect
from matplotlib.figure import Figure
from dac.core import DataNode, ContextKeyNode, ActionNode

import numpy as np
from time import sleep

class ActionBase(ActionNode): # needs thread
    QUICK_TASKS = []
    DEFAULT_TASK = None
    # the tasks to assist config input (e.g. browse files instead of filling manually)
    # NOTE: no thread for running the tasks, keep them simple

# TODO: add batch processing: read data, process with predefined parameters, process, save or export, clean and free memory

class ProcessActionBase(ActionBase):
    def __init__(self, context_key: ContextKeyNode, name: str = None, uuid: str = None) -> None:
        super().__init__(context_key, name, uuid)
        self._progress = lambda i, n: self._message(f"Progress: {i}/{n}")
        self._message = print

    def progress(self, i, n):
        self._progress(i, n)
    def message(self, s):
        self._message(f"[{self.name}]>{s}")

class VisualizeActionBase(ActionBase):
    def __init__(self, context_key: ContextKeyNode, name: str = None, uuid: str = None) -> None:
        super().__init__(context_key, name, uuid)
        self._figure: Figure = None
        self._cids = [] # never overwrite in __call__
        self._widgets = [] # to host evented widgets

        # on_callback to change the node params itself

    @property
    def canvas(self):
        return self._figure.canvas
    
    @property
    def figure(self):
        return self._figure
    
    @figure.setter
    def figure(self, fig: Figure):
        # when a figure assigned to visualization, need to clean the previous events and widgets (that have events)
        canvas = fig.canvas
        self._figure = fig
        if hasattr(canvas, "_cids"):
            for cid in canvas._cids:
                canvas.mpl_disconnect(cid)
        if hasattr(canvas, "_widgets"):
            for widget in canvas._widgets:
                widget.disconnect_events()
        canvas._cids = self._cids = []
        canvas._widgets = self._widgets = []
        fig.clf()

    def post_run(self, *args, **kwargs):
        self.canvas.draw_idle()
        # I have a doubt here, this is not called for SAB.
        # But canvas redrawed, widgets redrawed because of thread ends?

PAB = ProcessActionBase
VAB = VisualizeActionBase

class SimpleAction(ActionBase):
    CAPTION = "Simple action"

class SimpleGlobalAction(PAB):
    CAPTION = "Simple global action"

class Separator(ActionBase):
    CAPTION = "--- [Separator] ---"
    def __call__(self):
        pass

class SequenceActionBase(PAB, VAB):
    CAPTION = "Not implemented sequence"
    _SEQUENCE = []

    def __init_subclass__(cls, seq: list[type[ActionNode]]) -> None:
        cls._SEQUENCE = seq
        
        signatures = {}
        for act_type in seq:
            signatures[act_type.__name__] = act_type._SIGNATURE

        cls._SIGNATURE = signatures
    
    def __init__(self, context_key: ContextKeyNode, name: str = None, uuid: str = None) -> None:
        super().__init__(context_key, name, uuid)
        self._construct_config = SequenceActionBase._GetCCFromS(self._SIGNATURE)

    @staticmethod
    def _GetCCFromS(sig: inspect.Signature | dict):
        # get construct config from signature
        cfg = {}
        
        if isinstance(sig, inspect.Signature):
            for key, param in sig.parameters.items():
                if key=="self":
                    continue
                elif param.default is not inspect._empty:
                    cfg[key] = param.default
                elif param.annotation is not inspect._empty:
                    cfg[key] = ActionNode.Annotation2Config(param.annotation)
                else:
                    cfg[key] = "<Any>"
            if (ret_ann:=sig.return_annotation) is not inspect._empty and ret_ann.__name__!="list":
                ... # how to pass the return result?
        else:
            for subact_name, subact_sig in sig.items():
                cfg[subact_name] = SequenceActionBase._GetCCFromS(subact_sig)

        return cfg
    
    def __call__(self, **params):
        # using dict => each action type can be used only once
        
        n = len(self._SEQUENCE)
        rsts = []

        for i, subact_type in enumerate(self._SEQUENCE):
            name = subact_type.__name__

            def embed_progross(j, m):
                self.progress(i*m+j, n*m)
            def embed_message(s):
                self.message(s)

            subact_params = params.get(name)
            self.message(f"Processing [{name}]")

            action = subact_type(self.context_key)
            action.name = name
            if isinstance(action, VAB):
                action._figure = self.figure # avoid figure.setter
            if isinstance(action, PAB):
                action._progress = embed_progross
                action._message = embed_message

            # action.pre_run()
            rst = action(**subact_params)
            if isinstance(rst, DataNode):
                rsts.append(rst)
            elif isinstance(rst, list):
                rsts.extend(rst)
            else:
                ...
            if isinstance(action, VAB):
                self._cids.extend(action._cids)
                self._widgets.extend(action._widgets)
            # action.post_run()

            self.progress(i+1, n)

        # if len(rsts)==1:
        #     return rsts[0] # error, out_name is not defined
        # else:
        #     return rsts

        return rsts

SAB = SequenceActionBase
    
class A1(PAB):
    def __call__(self, sec: int=5, total: int=None):
        if total:
            for i in range(total):
                self.message(f"Starting ... {i+1}/{total}")
                sleep(sec)
                self.progress(i+1, total)
        else:
            self.message(f"Starting unknown.")
            sleep(sec)
class A2(VAB):
    def __call__(self, amp: float):
        ax = self.figure.gca()
        x = np.linspace(0, 4*np.pi, 400)
        ax.plot(x, amp*np.sin(x))
        sleep(amp)
class A1A2(SAB, seq=[A1, A2]):
    CAPTION = "Sequence action example"
class A1A2_2(SAB, seq=[A1A2, A1, A2]):
    CAPTION = "SAB x SAB"
    
class LoopActionBase(PAB):
    # for the case we want to run action multiple times with different parameters
    # original action output streamed/saved to outside
    def __init_subclass__(cls, action_type: type[ActionBase]) -> None:
        return super().__init_subclass__()
LAB = LoopActionBase

class RemoteProcessActionBase(PAB):
    # distribute the calculation (and container) to remote (cloud)
    ...


import numpy as np
import re
from scipy import signal

from dac.core.actions import ActionBase, VAB, PAB, SAB
from . import TimeData
from .data_loader import load_tdms
from ..nvh import FilterType

class LoadAction(PAB):
    CAPTION = "Load measurement data"
    def __call__(self, fpaths: list[str], ftype: str=None) -> list[TimeData]:
        n = len(fpaths)
        rst = []
        for i, fpath in enumerate(fpaths):
            if not fpath.upper().endswith("TDMS"):
                continue
            r = load_tdms(fpath=fpath)
            rst.extend(r)
            self.progress(i+1, n)
        return rst

class TruncAction(ActionBase):
    CAPTION = "Truncate TimeData"
    def __call__(self, channels: list[TimeData], duration: tuple[float, float]=[0, 0]) -> list[TimeData]:
        rst = []
        xfrom, xto = duration

        for i, channel in enumerate(channels):
            x = channel.x
            if xto==0:
                idx_to = None
            else:
                if xto<0:
                    xto = x[-1] + xto
                idx_to = np.searchsorted(x, xto)

            idx_from = np.searchsorted(x, xfrom)
            y = channel.y[idx_from:idx_to]
            rst.append(TimeData(f"{channel.name}-Trunc", y=y, dt=channel.dt, y_unit=channel.y_unit, comment=channel.comment))
        
        return rst

class FilterAction(ActionBase):
    CAPTION = "Filter TimeData"
    def __call__(self, channels: list[TimeData], freqs: tuple[float, float], order: int=3, filter_type: FilterType=FilterType.BandPass) -> list[TimeData]:
        """Smooth `TimeData` with `butter` filter.

        Parameters
        ----------
        channels : list[TimeData]
            Input channels, list of original `TimeData`
        freqs : (start_frequency, end_frequency)
        order : int
        filter_type : 'LowPass' | 'HighPass' | 'BandPass' | 'BandStop'

        Returns
        -------
        list[TimeData]
            A list of filtered `TimeData`

        Notes
        -----
        ...
        """
        # with current annotation mechanism, freq as single float won't be passed inside
        rst = []

        if filter_type in (FilterType.BandPass, FilterType.BandStop):
            w = np.array(freqs)
        else:
            w = freqs[0]

        for i, channel in enumerate(channels):
            Wn = w / (channel.fs / 2)
            b, a = signal.butter(order, Wn, filter_type.value)
            y = signal.filtfilt(b, a, channel.y)

            rst.append(TimeData(name=f"{channel.name}-FiltT", y=y, dt=channel.dt, y_unit=channel.y_unit, comment=channel.comment))

        return rst

class ResampleAction(ActionBase):
    CAPTION = "Resample data to"
    def __call__(self, channels: list[TimeData], dt: float=1) -> list[TimeData]:
        rst = []
        for i, channel in enumerate(channels):
            interval = int(dt // channel.dt)
            if interval > 1:
                rst.append(TimeData(name=channel.name, y=channel.y[::interval], dt=channel.dt*interval, y_unit=channel.y_unit, comment=channel.comment))
            else:
                rst.append(channel)
        return rst

class PrepDataAction(SAB, seq=[TruncAction, ResampleAction, FilterAction]): # example sequences
    ...

class EnvelopeTimeAction(PAB):
    CAPTION = "Envelop with Hilbert transform"
    # only works when there is positive and negative part
    def __call__(self, channels: list[TimeData], restore_offset: bool=False) -> list[TimeData]:
        rst = []
        for i, channel in enumerate(channels):
            channel_y = channel.y
            offset = np.mean(channel_y)
            analytic = signal.hilbert(channel_y-offset)
            env = np.abs(analytic)
            if restore_offset:
                env += offset

            rst.append(TimeData(name=f"{channel.name}-Env", y=env, dt=channel.dt, y_unit=channel.y_unit, comment=channel.comment))

        return rst

class ShowTimeDataAction(VAB):
    CAPTION = "Show measurement data"
    def __call__(self, channels: list[TimeData], plot_dt: float=None, xlim: tuple[float, float]=None, ylim: tuple[float, float]=None):
        fig = self.figure
        fig.suptitle("Time data visualization")

        ax = fig.gca()
        ax.set_xlabel("Time [s]")
        if xlim: ax.set_xlim(xlim)
        if ylim: ax.set_ylim(ylim)
        
        for channel in channels:
            x, y = channel.x, channel.y
            if plot_dt is not None:
                interval = int(plot_dt // channel.dt)
                if interval > 1:
                    x = x[::interval]
                    y = y[::interval]
            
            ax.plot(x, y, label=f"{channel.name} [{channel.y_unit}]")
        
        ax.legend(loc="upper right")

class CounterToTachoAction(ActionBase):
    CAPTION = "Encoder counter to tacho"
    def __call__(self, channel: TimeData, ppr: int=1024, sr_delta: float=0.1) -> TimeData:
        counter = channel.y
        delta = int(channel.fs * sr_delta)
        rpm = np.zeros(len(counter))
        rpm[delta:-delta] = (counter[2*delta:]-counter[:-2*delta]) / ppr / (2*delta*channel.dt / 60)
        rpm[:delta] = rpm[delta]
        rpm[-delta:] = rpm[-delta-1]

        return TimeData(name="Tacho", y=rpm, dt=channel.dt, y_unit="rpm")
    
class ChopOffSpikesAction(PAB):
    CAPTION = "Chop off spikes"
    def __call__(self, channels: list[TimeData], max_dydt: float, limits: tuple[float, float]=None) -> list[TimeData]:
        # `max_dydt`: maximum delta y per sec

        if limits is None:
            low_lim, high_lim = -np.inf, np.inf
        else:
            low_lim, high_lim = limits

        ret_channels = []
        n = len(channels)

        for j, channel in enumerate(channels):
            signal = channel
            orig_data = signal.y
            target_data = orig_data.copy()

            max_dy = max_dydt * signal.dt

            look4raise = True
            s, e = 0, 0

            if len(target_data)>0:
                pv = orig_data[0]
            else:
                continue
            
            for i, cv in enumerate(orig_data[1:]):
                dy = cv-pv
                if look4raise:
                    if dy > max_dy:
                        s = i # start from prev idx
                        sv = pv
                        look4raise = False
                else:
                    if np.abs(dy)<max_dy and np.abs(cv-sv)<max_dy*(i+1-s) and cv<=high_lim and cv>=low_lim:
                        look4raise = True
                        target_data[(s+1):(i+1)] = np.linspace(sv, cv, num=(i-s+2), endpoint=True)[1:-1]
                pv = cv
            
            ret_channels.append(
                TimeData(
                    name=f"{channel.name}-Chop",
                    y=target_data,
                    dt=signal.dt,
                    y_unit=signal.y_unit
                )
            )

            self.progress(j+1, n)

        return ret_channels

class PulseToAzimuthAction(ActionBase):
    CAPTION = "Pulse to azimuth"
    def __call__(self, channel: TimeData, ref_level: float, ppr: int=1, higher_as_pulse: bool=True, phase_shift: float=0, ref_channel: TimeData=None) -> TimeData:
        # `ref_channel` is for example when speed ramping, azimuth is not linear equal
        if ref_channel:
            sr_ratio = ref_channel.dt / channel.dt
            ref_data = ref_channel.y

        data = channel.y
        inpulse = data>ref_level if higher_as_pulse else data<ref_level
        indexes = np.arange(len(data))[inpulse]
        idx_diff = np.diff(indexes)
        # assert len(idx_diff) == len(indexes)-1
        idx_pulse_end = indexes[:-1][idx_diff>np.mean(idx_diff)]
        ang_data = np.zeros_like(data) * np.nan
        for i, (from_idx, to_idx) in enumerate(zip(idx_pulse_end[:-1], idx_pulse_end[1:])):
            if ref_channel:
                ref_indexes = (np.arange(from_idx, to_idx) / sr_ratio).astype(int)
                aligned_subdata = ref_data[ref_indexes]
                ang_data[from_idx:to_idx] = np.cumsum(aligned_subdata) / np.sum(aligned_subdata) * 360 + 360*i
            else:
                ang_data[from_idx:to_idx] = np.arange(to_idx-from_idx)/(to_idx-from_idx)*360 + 360*i
        
        return TimeData(
            name=f"Azi-{channel.name}", dt=channel.dt, y_unit="°",
            y=(ang_data/ppr+phase_shift)%360,
        )
    
class RefPulseToAzimuthAction(PAB):
    ... # create azimuth using reference

class OpAction(ActionBase):
    CAPTION = "Operation on TimeData"
    def __call__(self, channels: list[TimeData], op_str: str="{0}", y_unit: str="-") -> TimeData:
        # assert same dt and same length
        ys = [channel.y for channel in channels]
        y = eval(re.sub(r"{(\d+)}", r"ys[\1]", op_str))
        return TimeData("Channel-Op_ed", y=y, dt=channels[0].dt, y_unit=y_unit)


"""Snippet scope

This module provides a module scope for dynamically generated code.

In some scenarios, the analysis can be very temporary and limited to specific data.
It doesn't need to create complete new dac-modules.
So the analysis is also defined inside configuration file, under config['dac']['exec'].
The defined class will reside under this module `dac.core.snippet`.
"""

def exec_script(script: str):
    exec(script, globals=globals())


import importlib
import inspect
import json
import re
import sys
from functools import partial
from glob import glob
from io import BytesIO, StringIO
from os import path

import yaml
from matplotlib.backend_bases import key_press_handler
from matplotlib.backends.backend_qt5agg import (FigureCanvasQTAgg,
                                                NavigationToolbar2QT)
from matplotlib.figure import Figure
from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.Qsci import QsciLexerPython, QsciLexerYAML, QsciScintilla
from PyQt5.QtCore import Qt
from PyQt5.QtGui import QCloseEvent, QMouseEvent
from PyQt5.QtWidgets import (QMainWindow, QStyle, QTreeWidget, QTreeWidgetItem,
                             QWidget)

from importlib.metadata import version
from dac import APPNAME, PYPI_NAME
from dac.core import GCK, ActionNode, Container, DataNode, NodeBase, ContextKeyNode
from dac.core.actions import PAB, VAB, ActionBase
from dac.core.thread import ThreadWorker
from dac.core.plugin import use_plugin
from dac.gui.base import MainWindowBase
from dac.core.snippet import exec_script

NAME, TYPE, REMARK = range(3)
SET_RECENTDIR = "RecentDir"



class TaskBase:
    def __init__(self, dac_win: "MainWindow", name: str, *args):
        self.dac_win = dac_win
        self.name = name

    def request_update_action(self):
        pass

    def __call__(self, action: ActionBase):
        pass

class MainWindow(MainWindowBase):
    APPTITLE = APPNAME
    APPSETTING = QtCore.QSettings(APPNAME, "Main")

    def __init__(self) -> None:
        super().__init__()
        self.resize(1200, 800)

        self.figure: Figure = None

        self._create_ui()
        self._create_menu()
        self._create_status()
        self._route_signals()

        self.container: Container = None
        self.project_config_fpath: str = None
        self.exec_script: str = ""
        self.apply_config({})
        
    def _create_ui(self):
        super()._create_ui()

        self.setDockNestingEnabled(True)
        self.data_list_widget = data_list = DataListWidget(self)
        self.action_list_widget = action_list = ActionListWidget(self)
        self.node_editor = node_editor = NodeEditorWidget(self)

        data_list_docker = QtWidgets.QDockWidget("Data", self)
        data_list_docker.setWidget(data_list)
        action_list_docker = QtWidgets.QDockWidget("Action", self)
        action_list_docker.setWidget(action_list)
        node_editor_docker = QtWidgets.QDockWidget("Editor", self)
        node_editor_docker.setWidget(node_editor)

        self.figure = figure = Figure()
        self.canvas = canvas = FigureCanvasQTAgg(figure)
        self.navibar = navibar = NavigationToolbar2QT(canvas, self)

        canvas.mpl_connect("key_press_event", lambda event: key_press_handler(event, canvas, navibar))
        canvas.setFocusPolicy(Qt.FocusPolicy.StrongFocus)

        center_widget = QtWidgets.QWidget(self)
        vlayout = QtWidgets.QVBoxLayout(center_widget)
        vlayout.addWidget(canvas)
        vlayout.addWidget(navibar)

        self.setCentralWidget(center_widget)
        self.addDockWidget(Qt.DockWidgetArea.LeftDockWidgetArea, data_list_docker)
        self.addDockWidget(Qt.DockWidgetArea.LeftDockWidgetArea, node_editor_docker)
        self.splitDockWidget(data_list_docker, action_list_docker, Qt.Orientation.Horizontal)
    
    def _create_menu(self):
        super()._create_menu()
        
        menubar = self.menuBar()
        app_menu = menubar.addMenu("&App")

        new_project_action = app_menu.addAction("&New project")
        app_menu.addSeparator()
        save_project_action = app_menu.addAction("&Save project")
        saveas_project_action = app_menu.addAction("Save as ...")
        load_project_action = app_menu.addAction("&Load project")
        edit_exec_action = app_menu.addAction("&Edit exec script")
        app_menu.addSeparator()
        exit_action = app_menu.addAction("E&xit")

        def action_new_project():
            self.project_config_fpath = None
            self.apply_config({})
            self.message("New project created")
        def action_save():
            config_fpath = self.project_config_fpath
            if config_fpath is None:
                action_saveas()
                return
            with open(config_fpath, mode="w", encoding="utf8") as fp:
                config = self.get_config()
                json.dump(config, fp, indent=2)
                self.message(f"Save project to {config_fpath}")
        def action_saveas():
            fpath, fext = QtWidgets.QFileDialog.getSaveFileName(
                parent=self, caption="Save project configuration", filter="DAC config (*.dac.json);;All(*.*)",
                directory=self.APPSETTING.value(SET_RECENTDIR)
            )
            if not fpath:
                return
            self.APPSETTING.setValue(SET_RECENTDIR, path.dirname(fpath))
            self.project_config_fpath = fpath
            action_save()
            self.setWindowTitle(f"{path.basename(fpath)} | {self.APPTITLE}")
        def action_load_project():
            fpath, fext = QtWidgets.QFileDialog.getOpenFileName(
                parent=self, caption="Open project configuration", filter="DAC config (*.json);;All (*.*)",
                directory=self.APPSETTING.value(SET_RECENTDIR)
            )
            if not fpath:
                return
            self.APPSETTING.setValue(SET_RECENTDIR, path.dirname(fpath))
            with open(fpath, mode="r", encoding="utf8") as fp:
                config = json.load(fp)
            self.project_config_fpath = fpath
            self.apply_config(config)
            self.message(f"Project loaded from {fpath}")
        def action_edit_exec():
            script, ok = QtWidgets.QInputDialog.getMultiLineText(
                self, "Edit exec script", "Input Python snippet to be executed", self.exec_script
            )
            if not ok:
                return
            self.exec_script = script
            exec_script(script)

        new_project_action.triggered.connect(action_new_project)
        save_project_action.triggered.connect(action_save)
        saveas_project_action.triggered.connect(action_saveas)
        load_project_action.triggered.connect(action_load_project)
        edit_exec_action.triggered.connect(action_edit_exec)
        exit_action.triggered.connect(self.close)

        tool_menu = self._dac_menu
        copy_action = tool_menu.addAction("Copy figure", self.action_copy_figure)
        tool_menu.insertAction(tool_menu.actions()[0], copy_action)
        tool_menu.addAction("Reload modules", self.action_reload_modules, shortcut=Qt.CTRL+Qt.Key_R)

        # TODO: debug mode {no thread; show action code; send data to ipy; reload modules}

        menubar.addMenu(self._dac_menu)
    
    def _create_status(self):
        return super()._create_status()

    def _route_signals(self):
        self.data_list_widget.sig_edit_data_requested.connect(self.node_editor.edit_node)
        self.action_list_widget.sig_edit_action_requested.connect(self.node_editor.edit_node)
        self.data_list_widget.sig_action_update_requested.connect(
            self.action_list_widget.refresh
        )
        self.data_list_widget.sig_action_runall_requested.connect(
            self.action_list_widget.run_all_actions
        )
        self.action_list_widget.sig_data_update_requested.connect(
            self.data_list_widget.refresh
        )
        self.node_editor.sig_return_node.connect(self.data_list_widget.action_apply_node_config)
        self.node_editor.sig_return_node.connect(self.action_list_widget.action_apply_node_config)
        
    def action_copy_figure(self):
        if self.figure is None:
            return
        with BytesIO() as buf:
            self.figure.savefig(buf)
            QtWidgets.QApplication.clipboard().setImage(QtGui.QImage.fromData(buf.getvalue()))

    def action_reload_modules(self):
        dlg = MultipleItemsDialog(self, sorted([mod.__name__ for mod in Container._modules]),
                                  caption="Reload modules", label="Select modules which need reloaded. \nFor function dev purpose. \nIn some cases, reloading project is required to make the change effect.")
        dlg.exec()
        if dlg.result() and (itms := dlg.get_result()):
            for mod_name in itms:
                mod = importlib.import_module(mod_name)
                importlib.reload(mod)
                # the sequence may be an issue, e.g. ModA depend on ModB, but ModA get reloaded first, and then ModB. (Then reload it twice?)

    def spawn_cofigure(self):
        figure = Figure()
        canvas = FigureCanvasQTAgg(figure)
        navibar = NavigationToolbar2QT(canvas, self)

        canvas.mpl_connect("key_press_event", lambda event: key_press_handler(event, canvas, navibar))
        canvas.setFocusPolicy(Qt.FocusPolicy.StrongFocus)

        widget = QtWidgets.QWidget(self, flags=Qt.WindowType.Tool)
        widget.setWindowTitle("Copilot figure")
        widget.resize(650, 450)
        layout = QtWidgets.QVBoxLayout(widget)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(navibar)
        layout.addWidget(canvas)
        widget.show()

        return figure

    def use_plugin(self, setting_fpath: str, clean: bool=True):
        use_plugin(setting_fpath, clean, dac_win=self)
        self._plugin_menu.setTitle(f"Plugins: {path.basename(setting_fpath)}")

    def use_plugins_dir(self, setting_dpath: str, default: str=None):
        menubar = self.menuBar()
        self._plugin_menu = plugin_menu = menubar.addMenu("Plugins") # TODO: make it about-to-open style, so search files everytime

        def apply_plugin_file_gen(f):
            def apply_plugin_file():
                self.use_plugin(f, clean=True)
            return apply_plugin_file
        
        for f in glob(path.join(setting_dpath, "*.yaml")):
            act = plugin_menu.addAction(path.basename(f))
            act.triggered.connect(apply_plugin_file_gen(f))

        if default:
            self.use_plugin(path.join(setting_dpath, default))

    def apply_config(self, config: dict):
        if self.project_config_fpath:
            self.setWindowTitle(f"{path.basename(self.project_config_fpath)} | {self.APPTITLE}")
        else:
            self.setWindowTitle(f"[New project] | {self.APPTITLE}")

        dac_config: dict = config.get("dac", {})
        self.exec_script = script = dac_config.get("exec", "")
        if script:
            _, ok = QtWidgets.QInputDialog.getMultiLineText(
                self, "Embedded exec script found", "Following snippet found, are you sure to execute it?\nPlease make sure the codes are safe.",
                script
            )
            if ok:
                exec_script(script)
        self.container = container = Container.parse_save_config(dac_config)
        self.data_list_widget.refresh(container)
        self.action_list_widget.refresh(container)

    def get_config(self):
        container_config = {} if self.container is None else self.container.get_save_config()

        return {
            "dac": {
                "_": {"version": version(PYPI_NAME)},
                "exec": self.exec_script, # don't add it by default
                **container_config,
            }
        }

class DataListWidget(QTreeWidget):
    sig_edit_data_requested = QtCore.pyqtSignal(DataNode)
    sig_action_update_requested = QtCore.pyqtSignal()
    sig_action_runall_requested = QtCore.pyqtSignal()
    
    def __init__(self, parent: MainWindow) -> None:
        super().__init__(parent)
        self._STYLE = self.style()
        self.dac_win = parent
        self._container: Container = None

        self.setHeaderLabels(["Name", "Type", "Remark"])
        self.setColumnWidth(NAME, 150)
        self.setColumnWidth(TYPE, 200)

        self.setSelectionMode(self.SelectionMode.ExtendedSelection)
        self.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.customContextMenuRequested.connect(self.action_context_requested)
        self.itemClicked.connect(self.action_item_clicked)
        self.itemDoubleClicked.connect(self.action_item_dblclicked)

    def refresh(self, container: Container=None):
        self.clear()
        if container is None:
            container = self._container
            if container is None:
                return
        else:
            self._container = container

        context_item = QtWidgets.QTreeWidgetItem(self)
        context_item.setText(NAME, "N/A")
        context_item.setText(TYPE, "Context")
        context_item.setData(NAME, Qt.ItemDataRole.UserRole, GCK)
        for node_type, node_name, node_object in container.context_keys.NodeIter:
            itm = QtWidgets.QTreeWidgetItem(context_item)
            itm.setText(NAME, node_name)
            itm.setText(TYPE, node_type.__name__)
            itm.setText(REMARK, node_object.uuid)
            itm.setData(NAME, Qt.ItemDataRole.UserRole, node_object)
            if container.current_key is node_object:
                itm.setIcon(NAME, self._STYLE.standardIcon(QStyle.StandardPixmap.SP_CommandLink))
        context_item.setExpanded(True)

        data_item = QtWidgets.QTreeWidgetItem(self)
        data_item.setText(NAME, container.current_key.name)
        data_item.setText(TYPE, "Data")
        for node_type, node_name, node_object in container.CurrentContext.NodeIter:
            itm = QtWidgets.QTreeWidgetItem(data_item)
            itm.setText(NAME, node_name)
            itm.setText(TYPE, node_type.__name__)
            itm.setText(REMARK, node_object.uuid)
            itm.setData(NAME, Qt.ItemDataRole.UserRole, node_object)
            itm.setData(TYPE, Qt.ItemDataRole.UserRole, True) # mark as un-editable
        data_item.setExpanded(True)

    def action_context_requested(self, pos: QtCore.QPoint):
        if (container := self._container) is None:
            return
        itm = self.itemAt(pos)
        menu = QtWidgets.QMenu("DataMenu")

        if (not itm) or not (node_object := itm.data(NAME, Qt.ItemDataRole.UserRole)):
            return
        
        if getattr(node_object, "QUICK_ACTIONS", []):
            nodes = []
            for i in self.selectedItems():
                node = i.data(NAME, Qt.ItemDataRole.UserRole)
                # if type(node) is type(node_object): # or subclass?
                #     nodes.append(node)
                nodes.append(node)

            def cb_quickaction_gen(qat: tuple[type[ActionBase], str, dict], data_nodes: list[DataNode]):
                act_type, data_param_name, other_params = qat
                def cb_quickaction():
                    params = {data_param_name: data_nodes, **other_params}
                    act = act_type(context_key=container.current_key)
                    act.container = container
                    if isinstance(act, VAB):
                        act.figure = self.dac_win.figure
                    act.pre_run()
                    act(**params)
                    act.post_run()
                return cb_quickaction
            
            for qat in node_object.QUICK_ACTIONS:
                qat: tuple[type[ActionBase], str, dict]
                act_type, data_param_name, other_params = qat
                menu.addAction(act_type.CAPTION).triggered.connect(cb_quickaction_gen(qat, nodes))
            menu.addSeparator()

        def cb_pushnode_gen(key_object):
            def cb_pushnode():
                self.dac_win.action_toggle_ipy_widget(dac_node=key_object)

            return cb_pushnode
        
        if (uneditable:=itm.data(TYPE, Qt.ItemDataRole.UserRole)): # data nodes in local context
            menu.addAction("Push to IPy").triggered.connect(cb_pushnode_gen(node_object))
            # TODO: enable delete local object
            menu.exec(self.viewport().mapToGlobal(pos))
            return # stop here, no activate / delete
        
        if node_object is GCK:
            def cb_creation_gen(n_t: type[DataNode]):
                def cb_creation():
                    new_node = n_t(name="[New node]")
                    container.context_keys.add_node(new_node)
                    self.refresh()
                    self.sig_edit_data_requested.emit(new_node)
                return cb_creation
            
            for n_t in Container.GetGlobalDataTypes():
                if isinstance(n_t, str):
                    menu.addAction(n_t).setEnabled(False)
                else:
                    menu.addAction(n_t.__name__).triggered.connect(cb_creation_gen(n_t))
        else:
            def cb_activate_gen(key_object):
                def cb_activate():
                    container.activate_context(key_object)
                    self.sig_action_update_requested.emit()
                    self.refresh()
                return cb_activate
            
            def cb_activaterun_gen(key_object):
                def cb_activaterun():
                    cb_activate_gen(key_object)()
                    self.sig_action_runall_requested.emit()
                return cb_activaterun

            menu.addAction("Activate and Run-all").triggered.connect(cb_activaterun_gen(node_object))
            
            if node_object is container.current_key:
                menu.addAction("De-activate").triggered.connect(cb_activate_gen(GCK))
            else:
                menu.addAction("Activate").triggered.connect(cb_activate_gen(node_object))

            def cb_del_gen(key_object):
                def cb_del():
                    if key_object is container.current_key:
                        container.activate_context(GCK)
                        self.sig_action_update_requested.emit()

                    container.remove_context_key(key_object)
                    self.refresh()

                return cb_del
            
            menu.addSeparator()
            menu.addAction("Push to IPy").triggered.connect(cb_pushnode_gen(node_object))
            menu.addAction("Delete").triggered.connect(cb_del_gen(node_object))

        menu.exec(self.viewport().mapToGlobal(pos))

    def action_item_clicked(self, item: QTreeWidgetItem, col: int):
        data = item.data(NAME, Qt.ItemDataRole.UserRole)
        uneditable = False # item.data(TYPE, Qt.ItemDataRole.UserRole)

        if uneditable or not isinstance(data, DataNode) or data is GCK: # GCK not edit-able
            return
        self.sig_edit_data_requested.emit(data)

    def action_item_dblclicked(self, item: QTreeWidgetItem, col: int):
        uneditable = item.data(TYPE, Qt.ItemDataRole.UserRole)

        if uneditable or (container := self._container) is None or not (node_object := item.data(NAME, Qt.ItemDataRole.UserRole)):
            return
        
        def cb_activate(key_object):
            container.activate_context(key_object)
            self.sig_action_update_requested.emit()
            self.refresh()

        if node_object is container.current_key:
            cb_activate(GCK)
        else:
            cb_activate(node_object)

    def mousePressEvent(self, e: QMouseEvent) -> None:
        # mid-btn-click => copy name. mid-button-click won't trigger 'itemClicked'
        if (e.button()==Qt.MouseButton.MidButton) and (itm := self.itemAt(e.pos())):
            name = itm.text(NAME)
            QtWidgets.QApplication.clipboard().setText(name)
        return super().mousePressEvent(e)
    
    def action_apply_node_config(self, node: DataNode, config: dict, fire: bool=False):
        if not isinstance(node, DataNode):
            return
        
        # if node from previous container, seems works too
        if (new_name:=config.get("name")) and new_name!=node.name:
            self._container.context_keys.rename_node_to(node, new_name) # NOTE: error when applying from local nodes
        node.apply_construct_config(config)
        self.refresh()

class ActionListWidget(QTreeWidget):
    sig_edit_action_requested = QtCore.pyqtSignal(ActionNode)
    sig_data_update_requested = QtCore.pyqtSignal()
    
    PIXMAP = {
        ActionNode.ActionStatus.INIT: QStyle.StandardPixmap.SP_FileIcon,
        ActionNode.ActionStatus.CONFIGURED: QStyle.StandardPixmap.SP_FileDialogContentsView,
        ActionNode.ActionStatus.COMPLETE: QStyle.StandardPixmap.SP_DialogApplyButton,
        ActionNode.ActionStatus.FAILED: QStyle.StandardPixmap.SP_DialogCancelButton,
    }

    def __init__(self, parent: MainWindow) -> None:
        super().__init__(parent)
        self._STYLE = self.style()
        self.dac_win = parent
        self._container: Container = None

        self.setHeaderLabels(["Name", "Output", "Remark"])
        self.setColumnWidth(NAME, 200)
        self.setColumnWidth(TYPE, 150)
        self.setSelectionMode(self.SelectionMode.ExtendedSelection)

        self.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.customContextMenuRequested.connect(self.action_context_requested)
        self.itemClicked.connect(self.action_item_clicked)
        self.itemDoubleClicked.connect(self.action_item_dblclicked)

    def refresh(self, container: Container=None):
        self.clear()
        if container is None:
            container = self._container
            if container is None:
                return
        else:
            self._container = container

        for action in container.ActionsInCurrentContext:
            itm = QtWidgets.QTreeWidgetItem(self)
            itm.setText(NAME, action.name)
            itm.setData(NAME, Qt.ItemDataRole.UserRole, action)
            if action.out_name is not None:
                itm.setText(TYPE, action.out_name)

            itm.setIcon(NAME, self._STYLE.standardIcon(ActionListWidget.PIXMAP[action.status]))

    def run_action(self, action: ActionNode, complete_cb: callable=None):
        if (container := self._container) is None:
            return
        params = container.prepare_params_for_action(action._SIGNATURE, action._construct_config)

        def completed(rst):
            current_context = container.CurrentContext
            if isinstance(rst, DataNode):
                rst.name = action.out_name # what if out_name is None?
                current_context.add_node(rst)
                self.sig_data_update_requested.emit()
            elif isinstance(rst, list):
                for e_rst in rst:
                    e_rst: DataNode # cautious if e_rst is not DataNode
                    current_context.add_node(e_rst)
                self.sig_data_update_requested.emit()
            else:
                pass # no output or other type_of_data

            action.status = ActionNode.ActionStatus.COMPLETE # TODO: update accordingly
            self.refresh()
            if callable(complete_cb):
                complete_cb()

        action.container = container
        self.dac_win.message(f"[{action.name}]")

        if isinstance(action, VAB):
            action.figure = self.dac_win.figure

        if isinstance(action, PAB):
            def fn(p, progress_emitter, logger):
                action._progress = progress_emitter
                action._message = logger
                action.pre_run()
                rst = action(**p)
                action.post_run()
                return rst
            worker = ThreadWorker(fn=fn, caption=action.name, p=params)
            worker.signals.result.connect(completed)
            self.dac_win.start_thread_worker(worker)
        else:
            action.pre_run()
            rst = action(**params)
            action.post_run()
            completed(rst)

    def run_all_actions(self): # both `container.get_node_of_type` and `current_context.add_node(rst)` need current context
        if (container := self._container) is None:
            return
        
        def action_yield():
            for action in container.ActionsInCurrentContext:
                if isinstance(action, VAB):
                    continue
                yield action

        action_yielder = action_yield()
        
        def run_next_action():
            try:
                action = next(action_yielder)
                self.run_action(action, run_next_action)
            except StopIteration:
                pass

        run_next_action()
        
    def action_context_requested(self, pos: QtCore.QPoint):
        if (container := self._container) is None:
            return
        itms = self.selectedItems()
        menu = QtWidgets.QMenu("ActionMenu")

        def add_new_actions(menu: QtWidgets.QMenu, index: int=None):
            menu_stack = []
            def cb_creation_gen(a_t: type[ActionNode]):
                def cb_creation():
                    a: ActionBase = a_t(context_key=container.current_key)
                    if (task := a.DEFAULT_TASK) is not None:
                        task: TaskBase
                        task(a) # no `request_update_action` needed

                    if index is None:
                        container.actions.append(a)
                    else:
                        container.actions.insert(index, a)
                    self.refresh()
                    self.sig_edit_action_requested.emit(a)
                return cb_creation
            
            for a_t in container.ActionTypesInCurrentContext:
                if isinstance(a_t, str):
                    if a_t.endswith(">]"):
                        menu_stack.append(menu)
                        menu = menu.addMenu(a_t)
                    elif a_t.endswith("<]"):
                        menu = menu_stack.pop()
                    else:
                        menu.addAction(a_t).setEnabled(False)
                else:
                    menu.addAction(a_t.CAPTION).triggered.connect(cb_creation_gen(a_t))
                    
            modifiers = QtWidgets.QApplication.keyboardModifiers()
            if modifiers & QtCore.Qt.KeyboardModifier.ShiftModifier:
                def custom_callback():
                    cls_path, ok = QtWidgets.QInputDialog.getText(self, "Custom input", "Input custom 'lib.module.action' to create action.")
                    if not ok:
                        return
                    a_t = Container.GetClass(cls_path)
                    cb_creation_gen(a_t=a_t)()
                menu.addAction(">> Custom input <<").triggered.connect(custom_callback)
        if not itms:
            add_new_actions(menu)
        else:
            acts = [itm.data(NAME, Qt.ItemDataRole.UserRole) for itm in itms]

            if len(acts)==1:
                def cb_task_gen(task, act):
                    def request_update_action():
                        self.sig_edit_action_requested.emit(act)

                    def cb_task():
                        task.request_update_action = request_update_action
                        task(act)
                        request_update_action()
                    return cb_task
                act: ActionBase = acts[0]
                for task in act.QUICK_TASKS:
                    task: TaskBase
                    menu.addAction(task.name).triggered.connect(cb_task_gen(task, act))
                menu.addSeparator()
                submenu = menu.addMenu("Insert before")
                add_new_actions(submenu, container.actions.index(act))

            def cb_cp2_gen(aa: list[ActionNode], context_key: ContextKeyNode):
                def cb_cp2():
                    for oa in aa:
                        oac = oa.get_construct_config()

                        a_t = oa.__class__
                        a = a_t(context_key=context_key)

                        a.apply_construct_config(oac)

                        container.actions.append(a)

                    if context_key is container.current_key:
                        self.refresh() # if not copy to self, no need to refresh

                return cb_cp2
            
            def cb_mvaft_gen(aa: list[ActionNode], a: ActionNode):
                def cb_mvaft():
                    for oa in aa:
                        container.actions.remove(oa)
                    idx = container.actions.index(a)
                    
                    container.actions[idx+1:idx+1] = aa

                    self.refresh()

                return cb_mvaft

            def cb_del_gen(aa: list[ActionNode]):
                def cb_del():
                    for a in aa:
                        container.actions.remove(a)
                    self.refresh()
                    
                return cb_del
            
            if container.current_key is not GCK:
                cp2menu = menu.addMenu("Copy to")
                current_type = type(container.current_key)
                for node_type, node_name, node in container.context_keys.NodeIter:
                    if isinstance(node, current_type):
                        # only allow copying to context of same type
                        cp2menu.addAction(node_name).triggered.connect(
                            cb_cp2_gen(acts, node)
                        )

            mvb4menu = menu.addMenu("Move after")
            for oa in container.actions:
                if oa.context_key is container.current_key and oa not in acts:
                    mvb4menu.addAction(oa.name).triggered.connect(cb_mvaft_gen(acts, oa))

            # TODO: change to drag&drop, mime data using indexes
                    
            def cb_showcode_gen(a: ActionNode):
                def cb_showcode():
                    try:
                        src = inspect.getsource(a.__class__)
                    except ModuleNotFoundError: # in compiled program, no src code
                        self.dac_win.message("No src code available", log=False)
                        return
                    editor = QsciScintilla(self.dac_win)
                    editor.setWindowFlag(Qt.WindowType.Tool)
                    editor.resize(1200, 520)
                    lexer = QsciLexerPython(editor)
                    lexer.setFont(QtGui.QFont("Consolas"))
                    editor.setLexer(lexer)
                    editor.setUtf8(True)
                    editor.setAutoIndent(True)
                    # editor.setEolVisibility(True)
                    editor.setIndentationGuides(True)
                    editor.setTabWidth(4)
                    editor.setIndentationsUseTabs(False)
                    editor.setMarginType(1, QsciScintilla.NumberMargin)

                    editor.setWindowTitle(a.name)
                    editor.setText(src)
                    editor.show()
                return cb_showcode
            
            menu.addSeparator()
            menu.addAction("Show code").triggered.connect(cb_showcode_gen(acts[0]))
            menu.addAction("Delete").triggered.connect(cb_del_gen(acts))

        menu.exec(self.viewport().mapToGlobal(pos))
    
    def action_item_clicked(self, item: QTreeWidgetItem, col: int):
        act = item.data(NAME, Qt.ItemDataRole.UserRole)
        # if not isinstance(act, ActionNode):
        #     return
        self.sig_edit_action_requested.emit(act)

    def action_item_dblclicked(self, item: QTreeWidgetItem, col: int):
        self.run_action( item.data(NAME, Qt.ItemDataRole.UserRole) )

    def action_apply_node_config(self, node: ActionNode, config: dict, fire: bool=False):
        if not isinstance(node, ActionNode):
            return
        
        # if node from previous container, seems works too
        node.apply_construct_config(config)

        node.status = ActionNode.ActionStatus.CONFIGURED
        if fire:
            self.run_action(node)
            # refresh included in `run_action`
        else:
            self.refresh()

class NodeEditorWidget(QWidget):
    sig_return_node = QtCore.pyqtSignal(NodeBase, dict, bool)

    def __init__(self, parent: MainWindow):
        super().__init__(parent)

        vlayout = QtWidgets.QVBoxLayout(self)
        vlayout.setContentsMargins(0, 0, 0, 0)

        self.editor = editor = QsciScintilla(self)
        lexer = QsciLexerYAML(editor)
        lexer.setFont(QtGui.QFont("Consolas"))
        editor.setLexer(lexer)
        editor.setUtf8(True)
        editor.setAutoIndent(True)
        # editor.setEolVisibility(True)
        editor.setIndentationGuides(True)
        editor.setTabWidth(4)
        editor.setIndentationsUseTabs(False)
        editor.setMarginType(1, QsciScintilla.NumberMargin)

        btn_layout = QtWidgets.QHBoxLayout()
        apply_btn = QtWidgets.QToolButton(self)
        apply_btn.setText("✔")
        apply_btn.setToolTip("Apply config")
        fire_btn = QtWidgets.QToolButton(self)
        fire_btn.setText("🔥")
        fire_btn.setToolTip("Fire = apply + run")
        apply_btn.clicked.connect(self.action_apply)
        fire_btn.clicked.connect(partial(self.action_apply, fire=True))

        vlayout.addWidget(editor)
        btn_layout.addStretch(1)
        btn_layout.addWidget(apply_btn)
        btn_layout.addWidget(fire_btn)
        vlayout.addLayout(btn_layout)

        self._current_node = None

    def edit_node(self, node: NodeBase):
        s = yaml.dump(node.get_construct_config(), allow_unicode=True, sort_keys=False)
        self.editor.setText(s + "\n# " + type(node).__name__)
        self._current_node = node

    def action_apply(self, fire=True):
        if self._current_node is None:
            return
        config = yaml.load(StringIO(self.editor.text()), Loader=yaml.FullLoader)
        self.sig_return_node.emit(self._current_node, config, fire)


class MultipleItemsDialog(QtWidgets.QDialog):
    def __init__(self, parent: QWidget, items: list[str], caption: str, label: str) -> None:
        super().__init__(parent)
        self._items = items

        self.setWindowTitle(caption)
        self.item_list = item_list = QtWidgets.QListWidget(self)
        item_list.setSelectionMode(item_list.SelectionMode.ExtendedSelection)
        item_list.addItems(items)
        cap_label = QtWidgets.QLabel(label, self)

        layout = QtWidgets.QVBoxLayout(self)
        layout.addWidget(cap_label)
        layout.addWidget(item_list)
        btn_box = QtWidgets.QDialogButtonBox(QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel, self)
        layout.addWidget(btn_box)

        btn_box.accepted.connect(self.accept)
        btn_box.rejected.connect(self.reject)

    def get_result(self):
        return [itm.text() for itm in self.item_list.selectedItems()]

if __name__=="__main__":
    app = QtWidgets.QApplication(sys.argv)
    win = MainWindow()
    win.use_plugins_dir(path.join(path.dirname(__file__), "../plugins"), default="0.base.yaml")
    win.show()
    app.exit(app.exec())


import json, sys, click
from os import path

from PyQt5 import QtWidgets
from dac.gui import MainWindow

@click.command()
@click.option("--config-file", help="Configuration file to load")
@click.option("--plugin-file", help="YAML file for plugins")
def main(config_file: str, plugin_file: str):
    app = QtWidgets.QApplication(sys.argv)
    win = MainWindow()

    # add splash progress for module loading

    if config_file is not None:
        with open(config_file, mode="r") as fp:
            config = json.load(fp)
            win.apply_config(config)

    # setting_fpath = path.join(path.dirname(__file__), "..", "plugins/0.base.yaml")
    # win.use_plugin(setting_fpath)
    win.use_plugins_dir(path.join(path.dirname(__file__), "../plugins"), default="0.base.yaml")

    win.show()
    app.exit(app.exec())

if __name__=="__main__":
    main()


from uuid import uuid4
from collections import defaultdict
from typing import Any
from types import GenericAlias
import inspect, importlib
from enum import IntEnum, Enum

class NodeBase:
    def __init__(self, name: str=None, uuid: str=None) -> None:
        self._hash = None
        self._construct_config = {}

        self.name = name
        self._uuid = uuid or str(uuid4()) # _ to avoid shown in construct_config

    @property
    def uuid(self):
        return self._uuid

    def calc_hash(self) -> str:
        ...

    def get_hash(self, force_recalc=False):
        if self._hash is None or force_recalc:
            self._hash = self.calc_hash()
        return self._hash
    
    def get_construct_config(self) -> dict:
        raise NotImplementedError

    def apply_construct_config(self, construct_config: dict):
        raise NotImplementedError

    def get_save_config(self) -> dict:
        cls = self.__class__

        return {
            "_uuid_": self.uuid,
            "_class_": f"{cls.__module__}.{cls.__qualname__}",
            **self.get_construct_config(),
        }

class NodeNotFoundError(Exception):
    pass



class DataNode(NodeBase):
    BASICTYPES = (int, float, str, bool)
    @staticmethod
    def Value2BasicTypes(v):
        if type(v) in DataNode.BASICTYPES: # `isinstance` not enough because e.g. `np.float64` is subclass of `float`
            return v
        elif isinstance(v, (list, tuple)):
            return [DataNode.Value2BasicTypes(e) for e in v]
        elif isinstance(v, dict):
            return {k: DataNode.Value2BasicTypes(e) for k, e in v.items()}
        else:
            return f"<{type(v).__name__}>"
        
    @staticmethod
    def ContainsOnlyBasicTypes(v):
        if isinstance(v, str) and v.startswith("<") and v.endswith(">"):
            return False
        elif isinstance(v, (list, tuple)):
            for e in v:
                if not DataNode.ContainsOnlyBasicTypes(e):
                    return False
        elif isinstance(v, dict):
            for k, e in v.items():
                if not DataNode.ContainsOnlyBasicTypes(e):
                    return False
        return True

    def get_construct_config(self) -> dict:
        # _construct_config is same as __dict__
        return {
            k: DataNode.Value2BasicTypes(v)
            for k, v in self.__dict__.items()
            if not k.startswith("_")
        }
    
    def apply_construct_config(self, construct_config: dict):
        for k, v in construct_config.items():
            if k in self.__dict__ and DataNode.ContainsOnlyBasicTypes(v):
                self.__dict__[k] = v

class ContextKeyNode(DataNode):
    pass

class GlobalContextKey(ContextKeyNode):
    pass

GCK = GlobalContextKey("Global Context")



class ActionNode(NodeBase):
    CAPTION = "Not implemented action"
    _SIGNATURE = None

    def __init_subclass__(cls) -> None:
        cls._SIGNATURE = inspect.signature(cls.__call__)

    class ActionStatus(IntEnum):
        INIT = 0
        CONFIGURED = 1
        COMPLETE = 2
        FAILED = -1

    @staticmethod
    def Annotation2Config(ann: type | GenericAlias): # | UnionType
        if hasattr(ann, '_fields'): # namedtuple
            return [f"[{f}]" for f in ann._fields]
        elif isinstance(ann, GenericAlias): # ok: list[], tuple[]; nok: dict[], type[]
            return [
                ActionNode.Annotation2Config(t)
                for t in ann.__args__
            ]
        else:
            return f"<{ann.__name__}>"

    def __init__(self, context_key: DataNode, name: str = None, uuid: str = None) -> None:
        super().__init__(name=self.CAPTION, uuid=uuid)
        
        self.status = ActionNode.ActionStatus.INIT
        self.out_name = None

        self.context_key = context_key
        self.container: Container = None # for the actions require external resources, normally when context_key is GCK

    def __call__(self, *args: Any, **kwds: Any) -> Any:
        # annotations have to be specified; if there is 'list', `list[...]` must be used
        # output type also needs specified
        print(f"'{self.name}' called with {args} and {kwds}.")
    
    def get_construct_config(self) -> dict:
        if not self._construct_config: # init
            cfg = self._construct_config
            for key, param in self._SIGNATURE.parameters.items():
                if key=="self":
                    continue
                elif param.default is not inspect._empty:
                    if isinstance(param.default, Enum):
                        cfg[key] = param.default.name
                    else:
                        cfg[key] = param.default
                elif param.annotation is not inspect._empty:
                    cfg[key] = ActionNode.Annotation2Config(param.annotation)
                else:
                    cfg[key] = "<Any>"
            if (ret_ann:=self._SIGNATURE.return_annotation) is not inspect._empty and ret_ann.__name__!="list":
                self.out_name = f"<{ret_ann.__name__}>"
                
        com_config = {
            "name": self.name,
            **self._construct_config,
        }
        if self.out_name is not None:
            com_config["out_name"] = self.out_name

        return com_config
    
    def apply_construct_config(self, construct_config: dict):
        if "name" in construct_config:
            self.name = construct_config["name"]
            # del construct_config["name"]
        if "out_name" in construct_config:
            self.out_name = construct_config["out_name"]
            # del construct_config["out_name"]

        # TODO: validate the construct_config
        self._construct_config = construct_config

        # if self.status==ActionNode.ActionStatus.INIT:
        self.status = ActionNode.ActionStatus.CONFIGURED
    
    def get_save_config(self) -> dict:
        cfg = super().get_save_config()

        if self.context_key is not GCK:
            cfg["_context_"] = self.context_key.uuid

        return cfg
    
    def pre_run(self, *args, **kwargs):
        ...

    def post_run(self, *args, **kwargs):
        ...



class DataContext(dict[type[DataNode], dict[str, DataNode]]):
    def __init__(self, container: "Container") -> None:
        super().__init__()
        self._container = container
        self._uuid_dict = {} # {uuid: (node_type, name)} # don't store object to avoid ref

    @property
    def NodeIter(self) -> list[tuple[type[DataNode], str, DataNode]]:
        for node_type, nodes in self.items():
            for node_name, node in nodes.items():
                yield (node_type, node_name, node)

    def add_node(self, node: NodeBase):
        node_type = type(node)
        if node_type in self:
            self[node_type][node.name] = node # in global context, should delete original node related actions (potential error)
        else:
            self[node_type] = {node.name: node}
        self._uuid_dict[node.uuid] = (node_type, node.name)

    def get_node_of_type(self, node_name: str, node_type: type[NodeBase]) -> NodeBase:
        try:
            return self[node_type][node_name]
        except KeyError:
            return None
        
    def rename_node_to(self, node: NodeBase, new_name: str):
        try:
            del self[type(node)][node.name]
        except:
            pass
        node.name = new_name
        self.add_node(node)

    def get_node_by_uuid(self, uuid: str) -> NodeBase:
        node_type, node_name = self._uuid_dict[uuid]
        return self[node_type][node_name]

class Container:
    _key_types = [] # [ type[context_key_node] ]
    _action_types = defaultdict(list) # {type[context_key_node]:  [ type[action_node] ]}
    _type_agencies = {} # {type: handler}
    _modules = set()

    def __init__(self) -> None:
        self.actions: list[ActionNode] = []
        self.contexts: dict[ContextKeyNode, DataContext] = defaultdict(lambda: DataContext(self))
        self.context_keys = DataContext(self)
        self.current_key: ContextKeyNode = GCK

    @property
    def CurrentContext(self) -> DataContext:
        return self.contexts[self.current_key]
    
    @property
    def ActionsInCurrentContext(self) -> list[ActionNode]:
        return filter(lambda a: a.context_key is self.current_key, self.actions)
    
    def get_node_of_type_for(self, context_key: ContextKeyNode, node_name: str, node_type: type[NodeBase]) -> NodeBase | None:
        context = self.get_context(context_key)
        if node:=context.get_node_of_type(node_name, node_type):
            return node
        elif (context_key is not GCK) and (node:=self.contexts[GCK].get_node_of_type(node_name, node_type)):
            return node
        elif node:=self.context_keys.get_node_of_type(node_name, node_type):
            return node
        else:
            return None

    def get_node_of_type(self, node_name: str, node_type: type[NodeBase]) -> NodeBase | None:
        return self.get_node_of_type_for(self.current_key, node_name, node_type)

    def activate_context(self, context_key: ContextKeyNode) -> DataContext:
        self.current_key = context_key
        return self.CurrentContext

    def get_context(self, context_key: ContextKeyNode) -> DataContext:
        return self.contexts[context_key]
    
    def remove_context_key(self, context_key: ContextKeyNode):
        del self.context_keys[type(context_key)][context_key.name]
        if context_key in self.contexts:
            del self.contexts[context_key]
        self.actions = [action for action in self.actions if action.context_key is not context_key]

    def _get_value_of_annotation(self, ann: type | GenericAlias, pre_value: Any):
        if pre_value is None:
            return None
        elif isinstance(ann, GenericAlias): # move before `issubclass`, otherwise error `ann` is not type
            if ann.__name__=="list" and len(ann.__args__)==1:
                value = []
                for c in pre_value:
                    try:
                        v = self._get_value_of_annotation(ann.__args__[0], c)
                    except NodeNotFoundError:
                        continue
                    value.append(v)
            elif ann.__name__=="tuple":
                value = [self._get_value_of_annotation(a, c) for a, c in zip(ann.__args__, pre_value)]
            else:
                raise NotImplementedError
            
            return value
        elif issubclass(ann, Enum):
            if isinstance(pre_value, Enum): # from default
                return pre_value
            else: # str
                return ann[pre_value]
        elif issubclass(ann, DataNode): # and isinstance(pre_value, str)
            if (value:=self.get_node_of_type(node_name=pre_value, node_type=ann)) is None:
                raise NodeNotFoundError(f"Node '{pre_value}' of '{ann.__name__}' not found.")
            return value
        elif ann in Container._type_agencies:
            return Container._type_agencies[ann](pre_value)
        else:
            return pre_value
    
    def prepare_params_for_action(self, signature: inspect.Signature | dict, construct_config: dict) -> dict:
        params = {}
        if isinstance(signature, inspect.Signature):
            for key, param in signature.parameters.items():
                if key=="self":
                    continue
                value = construct_config.get(key, param.default)
                if value is inspect._empty:
                    # not provided and no default
                    raise Exception(f"Parameter '{key}' not provided.")

                params[key] = self._get_value_of_annotation(param.annotation, value)
        else: # signature is a dict, SAB
            for subact_name, subact_signature in signature.items():
                subact_config = construct_config.get(subact_name, {})
                params[subact_name] = self.prepare_params_for_action(subact_signature, subact_config)
        return params
    
    def get_save_config(self):
        return {
            "actions": [action.get_save_config() for action in self.actions],
            "contexts": [n_o.get_save_config() for n_t, n_n, n_o in self.context_keys.NodeIter],
            # add combo_action definitions, combo_action = action1 + action2 + action3 ...
            # able to define pre-defined parameters and user-input parameters
            # only available in current project, but can be saved as a template
        }

    @classmethod
    def parse_save_config(cls, config: dict) -> "Container":
        container = Container()
        nodes = {}

        g_nodes = config.get("contexts") or config.get("global_nodes") or []
        for data_config in g_nodes:
            cls_path = data_config['_class_']
            del data_config['_class_']
            uuid = data_config['_uuid_']
            del data_config['_uuid_']

            try:
                data_class: type[DataNode] = Container.GetClass(cls_path)
            except AttributeError: # not found
                continue # TODO: log
            data_node = data_class(name="[Default]", uuid=uuid)
            data_node.apply_construct_config(data_config)

            nodes[uuid] = data_node

            container.context_keys.add_node(data_node)

        actions = config.get("actions") or []
        for act_config in actions:
            cls_path = act_config['_class_']
            del act_config['_class_']
            uuid = act_config['_uuid_']
            del act_config['_uuid_']

            try:
                act_class: type[ActionNode] = Container.GetClass(cls_path)
            except AttributeError:
                continue # TODO: log

            if '_context_' in act_config:
                if act_config['_context_'] not in nodes:
                    continue
                context_key = nodes[act_config['_context_']]
                del act_config['_context_']
            else:
                context_key = GCK
            
            action_node = act_class(context_key=context_key, uuid=uuid)

            action_node.apply_construct_config(act_config)

            container.actions.append(action_node)

        return container

    @staticmethod
    def GetClass(class_path: str) -> type[NodeBase]:
        module_name, class_name = class_path.rsplit(".", 1)
        module = importlib.import_module(module_name)
        Container._modules.add(module)
        return getattr(module, class_name)

    @staticmethod
    def RegisterGlobalDataType(node_type: type[ContextKeyNode] | str):
        Container._key_types.append(node_type)

    @staticmethod
    def GetGlobalDataTypes() -> list[type[ContextKeyNode] | str]:
        return Container._key_types

    @staticmethod
    def RegisterContextAction(context_type: type[ContextKeyNode], action_type: type[ActionNode] | str):
        Container._action_types[context_type].append(action_type)

    @staticmethod
    def GetContextActionTypes(context_type: type[ContextKeyNode]) -> list[type[ActionNode] | str]:
        return Container._action_types[context_type]
    
    @staticmethod
    def RegisterGlobalContextAction(action_type: type[ActionNode] | str):
        Container.RegisterContextAction(GlobalContextKey, action_type)

    @property
    def ActionTypesInCurrentContext(self) -> list[type[ActionNode] | str]:
        context_type = type(self.current_key)
        return Container.GetContextActionTypes(context_type)
    
    @staticmethod
    def RegisterNodeTypeAgency(node_type: type, agent_func: callable):
        # # `agent_func` can take any input,
        # # normally take one var-representative string
        # # and output an object
        # def agent_func(arg: Any) -> Any:
        #     ...
        Container._type_agencies[node_type] = agent_func


from dac.core.data import DataBase
import numpy as np

class TimeData(DataBase):
    def __init__(self, name: str = None, uuid: str = None, y: np.ndarray=None, dt: float=1, y_unit: str="-", comment: str="") -> None:
        super().__init__(name, uuid)

        self.y = y if y is not None else np.array([])
        self.dt = dt
        self.y_unit = y_unit
        self.comment = comment
        # t0

    @property
    def fs(self):
        return 1/self.dt
    
    @property
    def length(self):
        return len(self.y)
    
    @property
    def x(self):
        return np.arange(self.length) * self.dt
    
    @property
    def t(self):
        return self.x # combine with t0
    
    def to_bins(self, df: float, overlap: float) -> np.ndarray:
        y = self.y
        batch_N = int( 1/df * self.fs )
        stride_N = int( batch_N * (1-overlap) )
        N_batches = (self.length-batch_N) // stride_N + 1
        stride, = y.strides
        assert N_batches > 0

        batches = np.lib.stride_tricks.as_strided(y, shape=(N_batches, batch_N), strides=(stride*stride_N, stride))
        # batches -= batches_mean
        # # the code above will cause problem, it's a `as_strided` mapping
        # # corresponding values are connected

        return batches
    
    def effective_value(self):
        return np.sqrt(np.mean(self.y**2))